/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * UniCall support 
 * 
 * Copyright (C) 2003, Steve Underwood
 * based on work Copyright (C) 1999, Mark Spencer
 *
 * Mark Spencer <markster@linux-support.net>
 * Steve Underwood <steveu@coppice.org>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <stdio.h>
#include <string.h>
#include <asterisk/lock.h>
#include <asterisk/channel.h>
#include <asterisk/channel_pvt.h>
#include <asterisk/config.h>
#include <asterisk/logger.h>
#include <asterisk/module.h>
#include <asterisk/pbx.h>
#include <asterisk/options.h>
#include <asterisk/file.h>
#include <asterisk/ulaw.h>
#include <asterisk/alaw.h>
#include <asterisk/callerid.h>
#include <asterisk/adsi.h>
#include <asterisk/cli.h>
#include <asterisk/cdr.h>
#include <asterisk/musiconhold.h>
#include <asterisk/say.h>
#include <asterisk/app.h>
#include <asterisk/dsp.h>
#include <asterisk/utils.h>
#include <asterisk/causes.h>
#include <sys/signal.h>
#include <sys/select.h>
#include <errno.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/zaptel.h>
#include <math.h>
#include <ctype.h>
#include <tonezone.h>

#include <unicall.h>

#include "../asterisk.h"

/* 
   XXX 
   XXX   We definitely need to lock the private structure in unicall_read and such 
   XXX  
 */

/*
    When a physical interface is not in the connected state, processing for that channel should be
    handled completely by the monitor code. Once a completed call connection exists, the call can
    be handled in other ways - usually by its owner. At this stage, proper monitoring for call
    disconnection is still required.
 */
 
/* Dialing plan - bundled packages of TON and NPI */
#define PRI_INTERNATIONAL_ISDN			((UC_TON_INTERNATIONAL << 4) | UC_NPI_E163_E164)
#define PRI_NATIONAL_ISDN			((UC_TON_NATIONAL << 4) | UC_NPI_E163_E164)
#define PRI_LOCAL_ISDN				((UC_TON_SUBSCRIBER << 4) | UC_NPI_E163_E164)
#define PRI_PRIVATE				((UC_TON_SUBSCRIBER << 4) | UC_NPI_PRIVATE)
#define PRI_UNKNOWN				((UC_TON_UNKNOWN << 4) | UC_NPI_UNKNOWN)

#define CHANNEL_PSEUDO -12

static char *desc = "Unified call processing (UniCall)";
static char *tdesc = "Unified call processing (UniCall) driver";

static char *type = "UniCall";
static char *typecompat = "roT";    /* Retain compatibility with chan_tor */
static char *config = "unicall.conf";

#define NUM_SPANS       32
#define RESET_INTERVAL  3600    /* How often (in seconds) to reset unused channels */

#define CHAN_PSEUDO    -2

static char context[AST_MAX_EXTENSION] = "default";

static char language[MAX_LANGUAGE] = "";
static char musicclass[MAX_LANGUAGE] = "";

static char *cur_protocol_class = NULL;
static char *cur_protocol_variant = NULL;
static char *cur_protocol_end = NULL;
static int cur_group = 0;
static int cur_callergroup = 0;
static int cur_pickupgroup = 0;

static int immediate = 0;
static int stripmsd = 0;
static int callreturn = 0;
static int threewaycalling = 0;
static int transfer = 0;
static int cancallforward = 0;
static float rxgain = 0.0;
static float txgain = 0.0;
static int echocancel = 0;
static int echocanbridged = 0;
static char accountcode[20] = "";

static char callerid[256] = "";
static int use_callerid = 1;

static char mailbox[AST_MAX_EXTENSION];

static int amaflags = 0;

static int adsi = 0;

static int minunused = 2;
static char idleext[AST_MAX_EXTENSION];
static char idledial[AST_MAX_EXTENSION];

static int usecnt = 0;
AST_MUTEX_DEFINE_STATIC(usecnt_lock);

/* Protect the interface list (of unicall_pvt's) */
AST_MUTEX_DEFINE_STATIC(iflock);

/* Protect the monitoring thread, so only one process can kill or start it, and not
   when it's doing something critical. */
AST_MUTEX_DEFINE_STATIC(monlock);

/* This is the thread for the monitor which checks for input on the channels
   which are not currently in use.  */
static pthread_t monitor_thread = 0;

/* Chunk size to read -- we use 20ms chunks to make things happy.  */   
#define READ_SIZE 160

#define MASK_AVAIL              (1 << 0)        /* Channel available for PRI use */
#define MASK_INUSE              (1 << 1)        /* Channel currently in use */

struct unicall_pvt;

struct unicall_uc
{
    pthread_t master;                           /* Thread of master */
    pthread_mutex_t lock;                       /* Mutex */
    char idleext[AST_MAX_EXTENSION];            /* Where to idle extra calls */
    char idlecontext[AST_MAX_EXTENSION];        /* What context to use for idle */
    char idledial[AST_MAX_EXTENSION];           /* What to dial before dumping */
    int minunused;                              /* Min # of channels to keep empty */
    int minidle;                                /* Min # of "idling" calls to keep active */
    int nodetype;                               /* Node type */
    int channels;                               /* Num of chans in span (31 or 24) */
    uc_t *uc;
    int debug;
    int fd;
    int up;
    int offset;
    int span;
    int chanmask[32];                           /* Channel status */
    int resetting;
    int resetchannel;
    time_t lastreset;
    struct unicall_pvt *pvt[32];                /* Member channel pvt structs */
    struct unicall_channel *chan[32];           /* Channels on each line */
};

#if 0
#define DEFAULT_PRI_DEBUG (PRI_DEBUG_Q931_DUMP | PRI_DEBUG_Q921_DUMP | PRI_DEBUG_Q921_RAW | PRI_DEBUG_Q921_STATE)
#else
#define DEFAULT_PRI_DEBUG 0
#endif

#define SUB_REAL        0            /* Active call */
#define SUB_CALLWAIT    1            /* Call-Waiting call on hold */
#define SUB_THREEWAY    2            /* Three-way call */

static char *subnames[] =
{
    "Real",
    "Callwait",
    "Threeway"
};

struct unicall_subchannel
{
    int zfd;
    struct ast_channel *owner;
    int chan;
    short buffer[AST_FRIENDLY_OFFSET/2 + READ_SIZE];
    int buffer_contents;
    struct ast_frame f;                         /* One frame for each channel.  How did this ever work before? */
    int needringing;
    int needanswer;
    int current_codec;
    int inthreeway;
    int curconfno;                              /* What conference we're currently in */
    char dtmfq[AST_MAX_EXTENSION];
};

#define CONF_USER_REAL          (1 << 0)
#define CONF_USER_THIRDCALL     (1 << 1)

#define MAX_SLAVES              4

static struct unicall_pvt
{
    pthread_mutex_t lock;
    struct ast_channel *owner;                  /* Our current active owner (if applicable) */
                                                /* Up to three channels can be associated with this call */
        
    struct unicall_subchannel sub_unused;       /* Just a safety precaution */
    struct unicall_subchannel subs[3];          /* Sub-channels */

    struct unicall_pvt *slaves[MAX_SLAVES];     /* Slave to us (follows our conferencing) */
    struct unicall_pvt *master;                 /* Master to us (we follow their conferencing) */
    int inconference;                           /* If our real should be in the conference */
    
    char *protocol_class;                       /* Signalling style */
    char *protocol_variant;                     /* Signalling style */
    int radio;                                  /* radio type */
    int firstradio;                             /* first radio flag */
    float rxgain;
    float txgain;
    struct unicall_pvt *next;                   /* Next channel in list */
    char context[AST_MAX_EXTENSION];
    char exten[AST_MAX_EXTENSION];
    char language[MAX_LANGUAGE];
    char musicclass[MAX_LANGUAGE];
    char callerid[AST_MAX_EXTENSION];
    char rdnis[AST_MAX_EXTENSION];
    int group;
    int law;
    int confno;                         /* Our conference */
    int confusers;                      /* Who is using our conference */
    int propconfno;                     /* Propagated conference number */
    int callgroup;
    int pickupgroup;
    int immediate;                      /* Answer before getting digits? */
    int channel;                        /* Channel Number */
    int span;                           /* Span number */
    int dialing;
    int use_callerid;                   /* Whether or not to use caller id on this channel */
    int callreturn;
    int stripmsd;
    int echocancel;
    int echocanbridged;
    int echocanon;
    int threewaycalling;
    int transfer;
    int digital;
    int outgoing;
    int dnd;
    ZT_DIAL_OPERATION dop;
    int destroy;
    int ignoredtmf;                
    int inalarm;
    int reserved;
    char accountcode[20];               /* Account code */
    int amaflags;                       /* AMA Flags */
    int adsi;
    int cancallforward;
    char call_forward[AST_MAX_EXTENSION];
    char mailbox[AST_MAX_EXTENSION];
    int onhooktime;
    int msgstate;
    struct ast_dsp *dsp;
    
    int confirmanswer;                  /* Wait for '#' to confirm answer */
    int distinctivering;                /* Which distinctivering to use */
    
    int faxhandled;                     /* Has a fax tone already been handled? */
    
    char mate;                          /* flag to say its in MATE mode */
    int dtmfrelax;		        /* whether to run in relaxed DTMF mode */
    uc_t *uc;
    int blocked;
    int sigcheck;
    int alreadyhungup;
    uc_crn_t crn;
    //uc_call_t *call;
} *iflist = NULL;

static int restore_gains(struct unicall_pvt *p);

static int restart_monitor(void);

static int unicall_bridge(struct ast_channel *c0, struct ast_channel *c1, int flags, struct ast_frame **fo, struct ast_channel **rc);

static int unicall_sendtext(struct ast_channel *c, char *text);

static int unicall_open(char *fn);

static int unicall_close(int fd);

/* translate between Unicall causes and ast */
static int hangup_uc2cause(int cause)
{
    switch (cause)
    {
    case UC_CAUSE_USER_BUSY:
        return AST_CAUSE_BUSY;
    case UC_CAUSE_NORMAL_CLEARING:
        return AST_CAUSE_NORMAL;
    case UC_CAUSE_NETWORK_CONGESTION:
    case UC_CAUSE_REQ_CHANNEL_NOT_AVAILABLE:
        return AST_CAUSE_CONGESTION;
    case UC_CAUSE_UNASSIGNED_NUMBER:
    case UC_CAUSE_NUMBER_CHANGED:
        return AST_CAUSE_UNALLOCATED;
    default:
        return AST_CAUSE_FAILURE;
    }
    return 0;
}

/* translate between ast cause and Unicall */
static int hangup_cause2uc(int cause)
{
    switch (cause)
    {
    case AST_CAUSE_BUSY:
	return UC_CAUSE_USER_BUSY;
    case AST_CAUSE_NORMAL:
    default:
        return UC_CAUSE_NORMAL_CLEARING;
    }
    return 0;
}

int select_codec(struct unicall_pvt *p, int index, int new_codec)
{
    int res;

    res = 0;
    if (p->subs[index].current_codec != new_codec)
    {
        switch (new_codec)
        {
        case AST_FORMAT_SLINEAR:
            if (uc_set_api_codec(p->uc, 0, UC_CODEC_LINEAR16))
                res = -1;
            /*endif*/
            break;
        case AST_FORMAT_ULAW:
            if (uc_set_api_codec(p->uc, 0, UC_CODEC_ULAW))
                res = -1;
            /*endif*/
            break;
        case AST_FORMAT_ALAW:
            if (uc_set_api_codec(p->uc, 0, UC_CODEC_ALAW))
                res = -1;
            /*endif*/
            break;
        default:
            return -1;
        }
        /*endswitch*/
        if (res)
            ast_log(LOG_WARNING, "Unable to set channel %d to codec 0x%X.\n", p->channel, new_codec);
        else
            p->subs[index].current_codec = new_codec;
        /*endif*/
    }
    /*endif*/
    return res;
}

static int unicall_get_index(struct ast_channel *ast, struct unicall_pvt *p, int nullok)
{
    int res;

    if (p->subs[SUB_REAL].owner == ast)
        res = SUB_REAL;
    else if (p->subs[SUB_CALLWAIT].owner == ast)
        res = SUB_CALLWAIT;
    else if (p->subs[SUB_THREEWAY].owner == ast)
        res = SUB_THREEWAY;
    else
    {
        res = -1;
        if (!nullok)
            ast_log(LOG_WARNING, "Unable to get index, and nullok is not asserted\n");
        /*endif*/
    }
    /*endif*/
    return res;
}

static void swap_subs(struct unicall_pvt *p, int a, int b)
{
    int tchan;
    int tinthreeway;
    struct ast_channel *towner;

    ast_log(LOG_DEBUG, "Swapping %d and %d\n", a, b);

    tchan = p->subs[a].chan;
    towner = p->subs[a].owner;
    tinthreeway = p->subs[a].inthreeway;

    p->subs[a].chan = p->subs[b].chan;
    p->subs[a].owner = p->subs[b].owner;
    p->subs[a].inthreeway = p->subs[b].inthreeway;

    p->subs[b].chan = tchan;
    p->subs[b].owner = towner;
    p->subs[b].inthreeway = tinthreeway;

    if (p->subs[a].owner)
        p->subs[a].owner->fds[0] = p->subs[a].zfd;
    /*endif*/
    if (p->subs[b].owner)
        p->subs[b].owner->fds[0] = p->subs[b].zfd;
    /*endif*/
}

static int alloc_sub(struct unicall_pvt *p, int x)
{
    ZT_BUFFERINFO bi;

    if (p->subs[x].zfd >= 0)
    {
        ast_log(LOG_WARNING, "%s subchannel of %d already in use\n", subnames[x], p->channel);
        return -1;
    }
    /*endif*/
    if ((p->subs[x].zfd = unicall_open("/dev/zap/pseudo")) < 0)
    {
        ast_log(LOG_WARNING, "Unable to open pseudo channel: %s\n", strerror(errno));
        return -1;
    }
    /*endif*/
    if (ioctl(p->subs[x].zfd, ZT_GET_BUFINFO, &bi))
    {
        ast_log(LOG_WARNING, "Unable to check buffer policy on channel %d\n", x);
    }
    else
    {
        bi.txbufpolicy = ZT_POLICY_IMMEDIATE;
        bi.rxbufpolicy = ZT_POLICY_IMMEDIATE;
        bi.numbufs = 4;
        bi.bufsize = READ_SIZE;
        if (ioctl(p->subs[x].zfd, ZT_SET_BUFINFO, &bi))
            ast_log(LOG_WARNING, "Unable to set buffer policy on channel %d\n", x);
        /*endif*/
    }
    /*endif*/
    if (ioctl(p->subs[x].zfd, ZT_CHANNO, &p->subs[x].chan))
    {
        ast_log(LOG_WARNING, "Unable to get channel number for pseudo channel on FD %d\n", p->subs[x].zfd);
        unicall_close(p->subs[x].zfd);
        p->subs[x].zfd = -1;
        return -1;
    }
    /*endif*/
    if (option_debug)
        ast_log(LOG_DEBUG, "Allocated %s subchannel on FD %d channel %d\n", subnames[x], p->subs[x].zfd, p->subs[x].chan);
    /*endif*/
    return 0;
}

static int unalloc_sub(struct unicall_pvt *p, int x)
{
    if (x == SUB_REAL)
    {
        ast_log(LOG_WARNING, "Trying to unalloc the real channel %d?!?\n", p->channel);
        return -1;
    }
    /*endif*/
    ast_log(LOG_DEBUG, "Released sub %d of channel %d\n", x, p->channel);
    if (p->subs[x].zfd < 0)
        unicall_close(p->subs[x].zfd);
    /*endif*/
    p->subs[x].zfd = -1;
    p->subs[x].chan = 0;
    p->subs[x].owner = NULL;
    p->subs[x].inthreeway = 0;
    p->subs[x].current_codec = 0;
    p->subs[x].curconfno = -1;
    return 0;
}

static int unicall_open(char *fn)
{
    int fd;
    int isnum;
    int chan;
    int x;
    struct zt_bufferinfo bi;

    chan = 0;
    isnum = 1;
    for (x = 0;  x < strlen(fn);  x++)
    {
        if (!isdigit(fn[x]))
        {
            isnum = 0;
            break;
        }
        /*endif*/
    }
    /*endfor*/
    if (isnum)
    {
        if ((chan = atoi(fn)) < 1)
        {
            ast_log(LOG_WARNING, "Invalid channel number '%s'\n", fn);
	    return -1;
	}
        /*endif*/
        fn = "/dev/zap/channel";
    }
    /*endif*/

    if ((fd = open(fn, O_RDWR | O_NONBLOCK)) < 0)
    {
        ast_log(LOG_WARNING, "Unable to open '%s': %s\n", fn, strerror(errno));
        return -1;
    }
    /*endif*/
    if (chan)
    {
        if (ioctl(fd, ZT_SPECIFY, &chan))
        {
            x = errno;
            close(fd);
            errno = x;
            ast_log(LOG_WARNING, "Unable to specify channel %d: %s\n", chan, strerror(errno));
            return -1;
        }
        /*endif*/
    }
    /*endif*/
    memset(&bi, 0, sizeof(bi));
    if (ioctl(fd, ZT_GET_BUFINFO, &bi))
    {
        x = errno;
        close(fd);
        errno = x;
        ast_log(LOG_WARNING, "Unable to check buffer policy on channel %d: %s\n", chan, strerror(errno));
        return -1;
    }
    /*endif*/
    bi.txbufpolicy = ZT_POLICY_IMMEDIATE;
    bi.rxbufpolicy = ZT_POLICY_IMMEDIATE;
    bi.numbufs = 4;
    bi.bufsize = READ_SIZE;
    if (ioctl(fd, ZT_SET_BUFINFO, &bi))
    {
        x = errno;
        close(fd);
        errno = x;
        ast_log(LOG_WARNING, "Unable to set buffer policy on channel %d: %s\n", chan, strerror(errno));
        return -1;
    }
    /*endif*/
    return fd;
}

static int unicall_close(int fd)
{
    return close(fd);
}

static int unicall_digit(struct ast_channel *ast, char digit)
{
    struct unicall_pvt *p;
    ZT_DIAL_OPERATION zo;
    int res = 0;
    int index;

    p = ast->pvt->pvt;
    if ((index = unicall_get_index(ast, p, 0)) == SUB_REAL)
    {
        zo.op = ZT_DIAL_OP_APPEND;
        zo.dialstr[0] = 'T';
        zo.dialstr[1] = digit;
        zo.dialstr[2] = 0;
        if ((res = ioctl(p->subs[SUB_REAL].zfd, ZT_DIAL, &zo)))
            ast_log(LOG_WARNING, "Couldn't dial digit %c\n", digit);
        else
            p->dialing = 1;
        /*endif*/
    }
    /*endif*/    
    return res;
}

#if 0
static char *zt_events[] =
{
    "No event",
    "On hook",
    "Ring/Answered",
    "Wink/Flash",
    "Alarm",
    "No more alarm",
    "HDLC Abort",
    "HDLC Overrun",
    "HDLC Bad FCS",
    "Dial Complete",
    "Ringer On",
    "Ringer Off",
    "Hook Transition Complete",
    "Bits Changed",
    "Pulse Start"
};
 
static char *zt_event2str(int event)
{
    static char buf[256];

    if (-1 < event  &&  event < 15)
        return zt_events[event];
    /*endif*/
    sprintf(buf, "Event %d", event);
    return buf;
}
#endif

static void unicall_message(char *s)
{
    ast_verbose(s);
}

static void unicall_error(char *s)
{
    ast_log(LOG_WARNING, "UniCall: %s", s);
}

static int conf_add(int *confno, struct unicall_subchannel *c, int index)
{
    ZT_CONFINFO zi;

    /* If the conference already exists, and we're already in it
       don't bother doing anything */
    if ((*confno > 0)  &&  (c->curconfno == *confno))
        return 0; 
    /*endif*/
    if (c->curconfno > 0)
        ast_log(LOG_WARNING, "Subchannel %d is already in conference %d, moving to %d\n", c->zfd, c->curconfno, *confno);
    /*endif*/
    if (c->zfd < 0)
        return 0;
    /*endif*/
    memset(&zi, 0, sizeof(zi));
    zi.chan = 0;
    zi.confno = *confno;
    if (index == SUB_REAL)
    {
        /* Real-side and pseudo-side both participate in conference */
        zi.confmode = ZT_CONF_REALANDPSEUDO | ZT_CONF_TALKER | ZT_CONF_LISTENER | ZT_CONF_PSEUDO_TALKER | ZT_CONF_PSEUDO_LISTENER;
    }
    else
    {
        zi.confmode = ZT_CONF_CONF | ZT_CONF_TALKER | ZT_CONF_LISTENER;
    }
    /*endif*/
    if (ioctl(c->zfd, ZT_SETCONF, &zi))
    {
        ast_log(LOG_WARNING, "Failed to add %d to conference %d\n", c->zfd, *confno);
        return -1;
    }
    /*endif*/
    c->curconfno = zi.confno;
    *confno = zi.confno;
    ast_log(LOG_DEBUG, "Added %d to conference %d\n", c->zfd, *confno);
    return 0;
}

static int conf_del(int *confno, struct unicall_subchannel *c, int index)
{
    ZT_CONFINFO zi;

    /* Can't delete from this conference if it's not 0 */
    /* Can't delete if there's no z */
    /* Don't delete from the conference if it's not our conference */
    /* Don't delete if we don't think it's conferenced at all (implied) */
    if (*confno < 1  ||  c->zfd < 0  ||  (*confno != c->curconfno))
        return 0;
    /*endif*/
    memset(&zi, 0, sizeof(zi));
    zi.chan = 0;
    zi.confno = 0;
    zi.confmode = 0;
    if (ioctl(c->zfd, ZT_SETCONF, &zi))
    {
        ast_log(LOG_WARNING, "Failed to drop %d from conference %d\n", c->zfd, *confno);
        return -1;
    }
    /*endif*/
    c->curconfno = -1;
    ast_log(LOG_DEBUG, "Removed %d from conference %d\n", c->zfd, *confno);
    return 0;
}

static inline int unicall_confmute(struct unicall_pvt *p, int muted)
{
    int x;
    int res;

    x = muted;
    if ((res = ioctl(p->subs[SUB_REAL].zfd, ZT_CONFMUTE, &x)) < 0) 
        ast_log(LOG_WARNING, "unicall_confmute(%d) failed on channel %d: %s\n", muted, p->channel, strerror(errno));
    /*endif*/
    return res;
}

static int update_conf(struct unicall_pvt *p)
{
    int needconf = 0;
    int x;

    /* Update conference state in a stateless fashion */
    /* Start with the obvious, general stuff */
    for (x = 0;  x < 3;  x++)
    {
        if (p->subs[x].zfd >= 0  &&  p->subs[x].inthreeway)
        {
            conf_add(&p->confno, &p->subs[x], x);
            needconf++;
        }
        else
        {
            conf_del(&p->confno, &p->subs[x], x);
        }
        /*endif*/
    }
    /*endfor*/
    /* If we have a slave, add him to our conference now */
    for (x = 0;  x < MAX_SLAVES;  x++)
    {
        if (p->slaves[x])
        {
            conf_add(&p->confno, &p->slaves[x]->subs[SUB_REAL], SUB_REAL);
            needconf++;
        }
        /*endif*/
    }
    /*endfor*/
    /* If we're supposed to be in there, do so now */
    if (p->inconference  &&  !p->subs[SUB_REAL].inthreeway)
    {
        conf_add(&p->confno, &p->subs[SUB_REAL], SUB_REAL);
        needconf++;
    }
    /*endif*/
    /* If we have a master, add ourselves to his conference */
    if (p->master) 
        conf_add(&p->master->confno, &p->subs[SUB_REAL], SUB_REAL);
    /*endif*/
    if (!needconf)
    {
        /* Nobody is left (or should be left) in our conference.  
           Kill it.  */
        p->confno = -1;
    }
    /*endif*/
    ast_log(LOG_DEBUG, "Updated conferencing on %d, with %d conference users\n", p->channel, needconf);
    return 0;
}

static void unicall_enable_ec(struct unicall_pvt *p)
{
    int x;
    int res;

    if (p == NULL)
        return;
    /*endif*/
    if (p->echocanon)
    {
        ast_log(LOG_DEBUG, "Echo cancellation already on\n");
        return;
    }
    /*endif*/
    if (p->echocancel)
    {
        x = p->echocancel;
        if ((res = ioctl(p->subs[SUB_REAL].zfd, ZT_ECHOCANCEL, &x)))
        {
            ast_log(LOG_WARNING, "Unable to enable echo cancellation on channel %d\n", p->channel);
        }
        else
        {
            p->echocanon = 1;
            ast_log(LOG_DEBUG, "Enabled echo cancellation on channel %d\n", p->channel);
        }
        /*endif*/
    }
    else
    {
        ast_log(LOG_DEBUG, "No echocancellation requested\n");
    }
    /*endif*/
}

static void unicall_disable_ec(struct unicall_pvt *p)
{
    int x;
    int res;

    if (p == NULL)
        return;
    /*endif*/
    if (p->echocancel)
    {
        x = 0;
        if ((res = ioctl(p->subs[SUB_REAL].zfd, ZT_ECHOCANCEL, &x)))
            ast_log(LOG_WARNING, "Unable to disable echo cancellation on channel %d\n", p->channel);
        else
	    ast_log(LOG_DEBUG, "disabled echo cancellation on channel %d\n", p->channel);
        /*endif*/
    }
    /*endif*/
    p->echocanon = 0;
}

int set_actual_gain(int fd, int chan, float rxgain, float txgain, int law)
{
    struct zt_gains g;
    float ltxgain;
    float lrxgain;
    int j;
    int k;

    g.chan = chan;
    if (rxgain != 0.0  ||  txgain != 0.0)
    {
        /* calculate linear value of tx gain */
        ltxgain = pow(10.0, txgain/20.0);
        /* calculate linear value of rx gain */
        lrxgain = pow(10.0, rxgain/20.0);
        if (law == UC_CODEC_ALAW)
        {
            for (j = 0;  j < 256;  j++)
            {
                k = (int) (((float) AST_ALAW(j))*lrxgain);
                if (k > 32767)
                    k = 32767;
                /*endif*/
                if (k < -32767)
                    k = -32767;
                /*endif*/
                g.rxgain[j] = AST_LIN2A(k);
                k = (int) (((float) AST_ALAW(j))*ltxgain);
                if (k > 32767)
                    k = 32767;
                /*endif*/
                if (k < -32767)
                    k = -32767;
                /*endif*/
                g.txgain[j] = AST_LIN2A(k);
            }
            /*endfor*/
        }
        else
        {
            for (j = 0;  j < 256;  j++)
            {
                k = (int) (((float) AST_MULAW(j))*lrxgain);
                if (k > 32767)
                    k = 32767;
                /*endif*/
                if (k < -32767)
                    k = -32767;
                /*endif*/
                g.rxgain[j] = AST_LIN2MU(k);
                k = (int) (((float) AST_MULAW(j))*ltxgain);
                if (k > 32767)
                    k = 32767;
                /*endif*/
                if (k < -32767)
                    k = -32767;
                /*endif*/
                g.txgain[j] = AST_LIN2MU(k);
            }
            /*endfor*/
        }
        /*endif*/
    }
    else
    {
        for (j = 0;  j < 256;  j++)
        {
            g.rxgain[j] = j;
            g.txgain[j] = j;
        }
        /*endfor*/
    }
    /*endif*/
    /* set 'em */
    return  ioctl(fd, ZT_SETGAINS, &g);
}

static int unicall_call(struct ast_channel *ast, char *rdest, int timeout)
{
    struct unicall_pvt *p = ast->pvt->pvt;
    uc_callparms_t *callparms;
    char callerid[256];
    char dest[256];
    char *s;
    char *c;
    char *n;
    char *l;

ast_log(LOG_WARNING, "unicall_call called - '%s'\n", rdest);
    if ((ast->_state != AST_STATE_DOWN)  &&  (ast->_state != AST_STATE_RESERVED))
    {
        ast_log(LOG_WARNING, "unicall_call called on %s, neither down nor reserved\n", ast->name);
        return -1;
    }
    /*endif*/

    if (p->radio)
    {
        /* if a radio channel, up immediately */
        /* Special pseudo -- automatically up */
        ast_setstate(ast, AST_STATE_UP); 
        return 0;
    }
    /*endif*/
    if ((callparms = uc_new_callparms(NULL)) == NULL)
        return -1;
    /*endif*/

    strncpy(dest, rdest, sizeof(dest) - 1);
    if ((c = strchr(rdest, '/')))
        c++;
    else
        c = dest;
    /*endif*/
    if (ast->callerid)
    {
        strncpy(callerid, ast->callerid, sizeof(callerid) - 1);
ast_log(LOG_WARNING, "unicall_call caller id - '%s'\n", callerid);
        ast_callerid_parse(callerid, &n, &l);
        if (l)
        {
            ast_shrink_phone_number(l);
            if (!ast_isphonenumber(l))
                l = NULL;
            /*endif*/
        }
        /*endif*/
    }
    else
    {
        l = NULL;
    }
    /*endif*/
    if (strlen(c) < p->stripmsd)
    {
        ast_log(LOG_WARNING, "Number '%s' is shorter than stripmsd (%d)\n", c, p->stripmsd);
        return -1;
    }
    /*endif*/
    p->dop.op = ZT_DIAL_OP_REPLACE;
    if ((s = strchr(c + p->stripmsd, 'w')))
    {
        snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "T%s", s);
        *s = '\0';
    }
    else
    {
        strcpy(p->dop.dialstr, "");
    }
    /*endif*/
#if 0
    uc_callparm_bear_cap_transfer_cap(callparms, int cap);
    uc_callparm_bear_cap_transfer_mode(callparms, int mode);
    uc_callparm_bear_cap_transfer_rate(callparms, int rate);
    uc_callparm_userinfo_layer1_protocol(callparms, int prot);
    uc_callparm_user_rate(callparms, int rate);
    uc_callparm_originating_ton(callparms, int ton);
    uc_callparm_originating_npi(callparms, int npi);

    uc_callparm_originating_sub_addr_ton(callparms, int ton);
    uc_callparm_originating_sub_addr_npi(callparms, int npi);
    uc_callparm_originating_sub_addr_number(callparms, char *num);
    uc_callparm_destination_npi(callparms, int npi);
    uc_callparm_destination_ton(callparms, int ton);

    uc_callparm_destination_sub_addr_npi(callparms, int npi);
    uc_callparm_destination_sub_addr_ton(callparms, int ton);
    uc_callparm_destination_sub_addr_number(callparms, char *num);

    uc_callparm_redirecting_cause(callparms, int cause);
    uc_callparm_redirecting_presentation(callparms, int pres);
    uc_callparm_redirecting_ton(callparms, int ton);
    uc_callparm_redirecting_npi(callparms, int npi);
    uc_callparm_redirecting_number(callparms, char *num);
#endif
    uc_callparm_originating_number(callparms, l);
    uc_callparm_destination_number(callparms, c + p->stripmsd);
    if (uc_MakeCall(p->uc, &(p->crn), callparms) != UC_RET_OK)
    {
        ast_log(LOG_WARNING, "uc_MakeCall() failed\n");
        return -1;
    }
    /*endif*/
    ast_setstate(ast, AST_STATE_DIALING);
    return 0;
}

static int destroy_channel(struct unicall_pvt *prev, struct unicall_pvt *cur, int now)
{
    int owned = 0;
    int i = 0;
    int res = 0;

    if (!now)
    {
        if (cur->owner)
            owned = 1;
        /*endif*/
        for (i = 0;  i < 3;  i++)
        {
            if (cur->subs[i].owner)
                owned = 1;
            /*endif*/
        }
        /*endfor*/
        if (!owned)
        {
            if (prev)
                prev->next = cur->next;
            else
                iflist = cur->next;
            /*endif*/
            if ((res = unicall_close(cur->subs[SUB_REAL].zfd)))
            {
                ast_log(LOG_ERROR, "Unable to close device on channel %d\n", cur->channel);
                free(cur);
                return -1;
            }
            /*endif*/
            free(cur);
        }
        /*endif*/
    }
    else
    {
        if (prev)
            prev->next = cur->next;
        else
            iflist = cur->next;
        /*endif*/
        if ((res = unicall_close(cur->subs[SUB_REAL].zfd)))
        {
            ast_log(LOG_ERROR, "Unable to close device on channel %d\n", cur->channel);
            free(cur);
            return -1;
        }
        /*endif*/
        free(cur);
    }
    /*endif*/
    return 0;
}

static int restore_gains(struct unicall_pvt *p)
{
    if (set_actual_gain(p->subs[SUB_REAL].zfd, 0, p->rxgain, p->txgain, p->law))
    {
        ast_log(LOG_WARNING, "Unable to restore gains: %s\n", strerror(errno));
        return -1;
    }
    /*endif*/
    return 0;
}

static int unicall_hangup(struct ast_channel *ast)
{
    struct unicall_pvt *p = ast->pvt->pvt;
    struct unicall_pvt *tmp;
    struct unicall_pvt *prev;
    int res;
    int index;
    int x;

ast_log(LOG_WARNING, "unicall_hangup(%s)\n", ast->name);
    if (option_debug)
        ast_log(LOG_DEBUG, "unicall_hangup(%s)\n", ast->name);
    /*endif*/
    if (p == NULL)
    {
        ast_log(LOG_WARNING, "Asked to hangup channel not connected\n");
        return 0;
    }
    /*endif*/
    index = unicall_get_index(ast, p, 1);

    restore_gains(p);
    if (p->dsp)
	ast_dsp_digitmode(p->dsp, DSP_DIGITMODE_DTMF | p->dtmfrelax);
    /*endif*/
    x = 0;
    unicall_confmute(p, 0);

    ast_log(LOG_DEBUG,
            "Hangup: channel: %d index = %d, normal = %d, callwait = %d, thirdcall = %d\n",
            p->channel,
            index,
            p->subs[SUB_REAL].zfd,
            p->subs[SUB_CALLWAIT].zfd,
            p->subs[SUB_THREEWAY].zfd);
    p->ignoredtmf = 0;
    
    if (index >= 0)
    {
        /* Real channel. Do some fixup */
        p->subs[index].owner = NULL;
        p->subs[index].needanswer = 0;
        p->subs[index].needringing = 0;
        p->subs[index].current_codec = 0;
        if (uc_set_api_codec(p->uc, 0, UC_CODEC_DEFAULT))
            ast_log(LOG_WARNING, "Unable to set law on channel %d to default\n", p->channel);
        /*endif*/
        switch (index)
        {
        case SUB_REAL:
            if (p->subs[SUB_CALLWAIT].zfd >= 0  &&  p->subs[SUB_THREEWAY].zfd >= 0)
            {
                ast_log(LOG_DEBUG, "Normal call hung up with both three way call and a call waiting call in place?\n");
                if (p->subs[SUB_CALLWAIT].inthreeway)
                {
                    /* We had flipped over to answer a callwait and now it's gone */
                    ast_log(LOG_DEBUG, "We were flipped over to the callwait, moving back and unowning.\n");
                    /* Move to the call-wait, but un-own us until they flip back. */
                    swap_subs(p, SUB_CALLWAIT, SUB_REAL);
                    unalloc_sub(p, SUB_CALLWAIT);
                    p->owner = NULL;
                }
                else
                {
                    /* The three way hung up, but we still have a call wait */
                    ast_log(LOG_DEBUG, "We were in the threeway and have a callwait still.  Ditching the threeway.\n");
                    swap_subs(p, SUB_THREEWAY, SUB_REAL);
                    unalloc_sub(p, SUB_THREEWAY);
                    if (p->subs[SUB_REAL].inthreeway)
                    {
                        /* This was part of a three way call.  Immediately make way for
                           another call */
                        ast_log(LOG_DEBUG, "Call was complete, setting owner to former third call\n");
                        p->owner = p->subs[SUB_REAL].owner;
                    }
                    else
                    {
                        /* This call hasn't been completed yet...  Set owner to NULL */
                        ast_log(LOG_DEBUG, "Call was incomplete, setting owner to NULL\n");
                        p->owner = NULL;
                    }
                    /*endif*/
                    p->subs[SUB_REAL].inthreeway = 0;
                }
                /*endif*/
            }
            else if (p->subs[SUB_CALLWAIT].zfd >= 0)
            {
                /* Move to the call-wait and switch back to them. */
                swap_subs(p, SUB_CALLWAIT, SUB_REAL);
                unalloc_sub(p, SUB_CALLWAIT);
                p->owner = p->subs[SUB_REAL].owner;
                if (p->subs[SUB_REAL].owner->bridge)
                    ast_moh_stop(p->subs[SUB_REAL].owner->bridge);
                /*endif*/
            }
            else if (p->subs[SUB_THREEWAY].zfd >= 0)
            {
                swap_subs(p, SUB_THREEWAY, SUB_REAL);
                unalloc_sub(p, SUB_THREEWAY);
                if (p->subs[SUB_REAL].inthreeway)
                {
                    /* This was part of a three way call.  Immediately make way for
                       another call */
                    ast_log(LOG_DEBUG, "Call was complete, setting owner to former third call\n");
                    p->owner = p->subs[SUB_REAL].owner;
                }
                else
                {
                    /* This call hasn't been completed yet...  Set owner to NULL */
                    ast_log(LOG_DEBUG, "Call was incomplete, setting owner to NULL\n");
                    p->owner = NULL;
                }
                /*endif*/
                p->subs[SUB_REAL].inthreeway = 0;
            }
            /*endif*/
            break;
        case SUB_CALLWAIT:
            /* Ditch the holding callwait call, and immediately make it availabe */
            if (p->subs[SUB_CALLWAIT].inthreeway)
            {
                /* This is actually part of a three way, placed on hold.  Place the third part
                   on music on hold now */
                if (p->subs[SUB_THREEWAY].owner  &&  p->subs[SUB_THREEWAY].owner->bridge)
                    ast_moh_start(p->subs[SUB_THREEWAY].owner->bridge, NULL);
                /*endif*/
                p->subs[SUB_THREEWAY].inthreeway = 0;
                /* Make it the call wait now */
                swap_subs(p, SUB_CALLWAIT, SUB_THREEWAY);
                unalloc_sub(p, SUB_THREEWAY);
            }
            else
            {
                unalloc_sub(p, SUB_CALLWAIT);
            }
            /*endif*/
            break;
        case SUB_THREEWAY:
            if (p->subs[SUB_CALLWAIT].inthreeway)
            {
                /* The other party of the three way call is currently in a call-wait state.
                   Start music on hold for them, and take the main guy out of the third call */
                if (p->subs[SUB_CALLWAIT].owner && p->subs[SUB_CALLWAIT].owner->bridge)
                    ast_moh_start(p->subs[SUB_CALLWAIT].owner->bridge, NULL);
                /*endif*/
                p->subs[SUB_CALLWAIT].inthreeway = 0;
            }
            /*endif*/
            p->subs[SUB_REAL].inthreeway = 0;
            /* If this was part of a three way call index, let us make
               another three way call */
            unalloc_sub(p, SUB_THREEWAY);
            break;
        default:
            /* This wasn't any sort of call, but how are we an index? */
            ast_log(LOG_WARNING, "Index found but not any type of call?\n");
            break;
        }
        /*endswitch*/
    }
    /*endif*/

    if (!p->subs[SUB_REAL].owner  &&  !p->subs[SUB_CALLWAIT].owner  &&  !p->subs[SUB_THREEWAY].owner)
    {
        p->owner = NULL;
        p->distinctivering = 0;
        p->confirmanswer = 0;
        p->outgoing = 0;
        p->digital = 0;
        p->faxhandled = 0;
        p->onhooktime = time(NULL);
        if (p->dsp)
        {
	    ast_dsp_free(p->dsp);
	    p->dsp = NULL;
	}
        /*endif*/
        if (uc_set_api_codec(p->uc, 0, UC_CODEC_DEFAULT))
            ast_log(LOG_WARNING, "Unable to set law on channel %d to default\n", p->channel);
        /*endif*/
        /* Perform low level drop call if no owner left */
	if (p->crn)
        {
ast_log(LOG_WARNING, "Causes %d %d\n", ast->hangupcause, hangup_cause2uc(ast->hangupcause));
            if (uc_DropCall(p->uc, p->crn, hangup_cause2uc(ast->hangupcause)) < 0)
		ast_log(LOG_WARNING, "uc_DropCall failed\n");
            /*endif*/
	    if (p->alreadyhungup)
		p->alreadyhungup = 0;
	    /*endif*/
	}
        else
        {
	    res = 0;
        }
        /*endif*/
        tone_zone_play_tone(p->subs[SUB_REAL].zfd, -1);
        x = 0;
        //ast_channel_setoption(ast, AST_OPTION_TONE_VERIFY, &x, sizeof(char), 0);
        //ast_channel_setoption(ast, AST_OPTION_TDD, &x, sizeof(char), 0);
        p->dialing = 0;
        strcpy(p->rdnis, "");
        update_conf(p);
        restart_monitor();
    }
    /*endif*/
    ast->pvt->pvt = NULL;
    ast_setstate(ast, AST_STATE_DOWN);
    ast_mutex_lock(&usecnt_lock);
    if (--usecnt < 0) 
        ast_log(LOG_WARNING, "Usecnt < 0???\n");
    /*endif*/
    ast_mutex_unlock(&usecnt_lock);
    ast_update_use_count();
    if (option_verbose > 2) 
        ast_verbose( VERBOSE_PREFIX_3 "Hungup '%s'\n", ast->name);
    /*endif*/
    ast_mutex_lock(&iflock);
    if (p->destroy)
    {
        for (prev = NULL, tmp = iflist;  iflist;  prev = tmp, tmp = tmp->next)
        {
            if (tmp == p)
            {
                destroy_channel(prev, tmp, 0);
                break;
            }
            /*endif*/
        }
        /*endfor*/
    }
    /*endif*/
    ast_mutex_unlock(&iflock);
    return 0;
}

static int unicall_answer(struct ast_channel *ast)
{
    struct unicall_pvt *p = ast->pvt->pvt;
    int index;
    int res;

    ast_setstate(ast, AST_STATE_UP);
    if ((index = unicall_get_index(ast, p, 0)) < 0)
        index = SUB_REAL;
    /*endif*/
    /* Nothing to do if a radio channel */
    if (p->radio)
        return 0;
    /*endif*/
ast_log(LOG_WARNING, "uc_AnswerCall\n");
    if ((res = uc_AnswerCall(p->uc, p->crn)))
        ast_log(LOG_WARNING, "Answer call failed :( on %s, cause %d\n", ast->name, res);
    /*endif*/
    return res;
}

static int unicall_setoption(struct ast_channel *chan, int option, void *data, int datalen)
{
    struct unicall_pvt *p = chan->pvt->pvt;
    char *cp;
    int x;

    ast_log(LOG_WARNING, "unicall_setoption called - %d\n", option);
    if (option != AST_OPTION_TONE_VERIFY
        &&
        option != AST_OPTION_TDD
        &&
        option != AST_OPTION_RELAXDTMF)
    {
        errno = ENOSYS;
        return -1;
    }
    /*endif*/
    if (data == NULL  ||  datalen < 1)
    {
        errno = EINVAL;
        return -1;
    }
    /*endif*/
    cp = (char *) data;
    switch (option)
    {
    case AST_OPTION_TONE_VERIFY:
        switch (*cp)
        {
        case 1:
            ast_log(LOG_DEBUG, "Set option TONE VERIFY, mode: MUTECONF(1) on %s\n", chan->name);
            ast_dsp_digitmode(p->dsp, DSP_DIGITMODE_MUTECONF | p->dtmfrelax);  /* set mute mode if desired */
            break;
        case 2:
            ast_log(LOG_DEBUG, "Set option TONE VERIFY, mode: MUTECONF/MAX(2) on %s\n", chan->name);
            ast_dsp_digitmode(p->dsp, DSP_DIGITMODE_MUTECONF | DSP_DIGITMODE_MUTEMAX | p->dtmfrelax);  /* set mute mode if desired */
            break;
        default:
            ast_log(LOG_DEBUG, "Set option TONE VERIFY, mode: OFF(0) on %s\n", chan->name);
            ast_dsp_digitmode(p->dsp, DSP_DIGITMODE_DTMF | p->dtmfrelax);  /* set mute mode if desired */
            break;
        }
        /*endswitch*/
        break;
    case AST_OPTION_RELAXDTMF:
        /* Relax DTMF decoding (or not) */
        switch (*cp)
        {
        case 0:
            ast_log(LOG_DEBUG, "Set option RELAX DTMF, value: OFF(0) on %s\n",chan->name);
            x = 0;
            break;
        default:
            ast_log(LOG_DEBUG, "Set option RELAX DTMF, value: ON(1) on %s\n",chan->name);
            x = 1;
            break;
        }
        /*endswitch*/
        ast_dsp_digitmode(p->dsp,x  ?  DSP_DIGITMODE_RELAXDTMF  :  (DSP_DIGITMODE_DTMF | p->dtmfrelax));
        break;
    }
    /*endswitch*/
    errno = 0;
    return 0;
}

static void unicall_unlink(struct unicall_pvt *slave, struct unicall_pvt *master)
{
    /* Unlink a specific slave or all slaves/masters from a given master */
    int x;
    int hasslaves;

    if (master == NULL)
        return;
    /*endif*/
    hasslaves = 0;
    for (x = 0;  x < MAX_SLAVES;  x++)
    {
        if (master->slaves[x])
        {
            if (slave == NULL  ||  master->slaves[x] == slave)
            {
                /* Take slave out of the conference */
                ast_log(LOG_DEBUG, "Unlinking slave %d from %d\n", master->slaves[x]->channel, master->channel);
                conf_del(&master->confno, &master->slaves[x]->subs[SUB_REAL], SUB_REAL);
                master->slaves[x]->master = NULL;
                master->slaves[x] = NULL;
            }
            else
            {
                hasslaves = 1;
            }
            /*endif*/
        }
        /*endif*/
        if (!hasslaves)
            master->inconference = 0;
        /*endif*/
    }
    /*endfor*/
    if (slave == NULL)
    {
        if (master->master)
        {
            /* Take master out of the conference */
            conf_del(&master->master->confno, &master->subs[SUB_REAL], SUB_REAL);
            hasslaves = 0;
            for (x = 0;  x < MAX_SLAVES;  x++)
            {
                if (master->master->slaves[x] == master)
                    master->master->slaves[x] = NULL;
                else if (master->master->slaves[x])
                    hasslaves = 1;
                /*endif*/
            }
            /*endfor*/
            if (!hasslaves)
                master->master->inconference = 0;
            /*endif*/
        }
        /*endif*/
        master->master = NULL;
    }
    /*endif*/
    update_conf(master);
}

static void unicall_link(struct unicall_pvt *slave, struct unicall_pvt *master)
{
    int x;

    if (slave == NULL  ||  master == NULL)
    {
        ast_log(LOG_WARNING, "Tried to link to/from NULL??\n");
        return;
    }
    /*endif*/
    for (x = 0;  x < MAX_SLAVES;  x++)
    {
        if (!master->slaves[x])
        {
            master->slaves[x] = slave;
            break;
        }
        /*endif*/
    }
    /*endfor*/
    if (x >= MAX_SLAVES)
    {
        ast_log(LOG_WARNING, "Replacing slave %d with new slave, %d\n", master->slaves[MAX_SLAVES - 1]->channel, slave->channel);
        master->slaves[MAX_SLAVES - 1] = slave;
    }
    /*endif*/
    if (slave->master) 
        ast_log(LOG_WARNING, "Replacing master %d with new master, %d\n", slave->master->channel, master->channel);
    /*endif*/
    slave->master = master;
    
    ast_log(LOG_DEBUG, "Making %d slave to master %d\n", slave->channel, master->channel);
}

static int unicall_bridge(struct ast_channel *c0, struct ast_channel *c1, int flags, struct ast_frame **fo, struct ast_channel **rc)
{
    struct unicall_pvt *p0;
    struct unicall_pvt *p1;
    struct unicall_pvt *op0;
    struct unicall_pvt *op1;
    struct unicall_pvt *master = NULL;
    struct unicall_pvt *slave = NULL;
    struct ast_channel *who = NULL;
    struct ast_channel *cs[3];
    struct ast_frame *f;
    int to;
    int inconf = 0;
    int nothingok = 0;
    int ofd1;
    int ofd2;
    int oi1;
    int oi2;
    int i1 = -1;
    int i2 = -1;
    int t1;
    int t2;
    int os1 = -1;
    int os2 = -1;
    struct ast_channel *oc1;
    struct ast_channel *oc2;

    ast_log(LOG_WARNING, "unicall_bridge called\n");
    /* if need DTMF, can't native bridge */
    if ((flags & (AST_BRIDGE_DTMF_CHANNEL_0 | AST_BRIDGE_DTMF_CHANNEL_1)))
        return -2;
    /*endif*/
    p0 = c0->pvt->pvt;
    p1 = c1->pvt->pvt;
    /* can't do pseudo-channels here */
    if (p0->protocol_class == NULL  ||  p1->protocol_class == NULL)
        return -2;
    /*endif*/
    ast_mutex_lock(&c0->lock);
    ast_mutex_lock(&c1->lock);
    op0 =
    p0 = c0->pvt->pvt;
    op1 =
    p1 = c1->pvt->pvt;
    ofd1 = c0->fds[0];
    ofd2 = c1->fds[0];
    oi1 = unicall_get_index(c0, p0, 0);
    oi2 = unicall_get_index(c1, p1, 0);
    oc1 = p0->owner;
    oc2 = p1->owner;
    if (oi1 < 0  ||  oi2 < 0)
        return -1;
    /*endif*/
    ast_mutex_lock(&p0->lock);
    if (pthread_mutex_trylock(&p1->lock))
    {
        /* Don't block, due to potential for deadlock */
        ast_mutex_unlock(&p0->lock);
        ast_mutex_unlock(&c0->lock);
        ast_mutex_unlock(&c1->lock);
        ast_log(LOG_NOTICE, "Avoiding deadlock...\n");
        return -3;
    }
    /*endif*/
    if (oi1 == SUB_REAL  &&  oi2 == SUB_REAL)
    {
        if (p0->owner == NULL  ||  p1->owner == NULL)
        {
            /* Currently unowned -- Do nothing. */
            nothingok = 1;
        }
        else
        {
            /* If we don't have a call-wait in a 3-way, and we aren't in a 3-way, we can be master */
            if (!p0->subs[SUB_CALLWAIT].inthreeway  &&  !p1->subs[SUB_REAL].inthreeway)
            {
                master = p0;
                slave = p1;
                inconf = 1;
            }
            else if (!p1->subs[SUB_CALLWAIT].inthreeway  &&  !p0->subs[SUB_REAL].inthreeway)
            {
                master = p1;
                slave = p0;
                inconf = 1;
            }
            else
            {
                ast_log(LOG_WARNING, "Huh?  Both calls are callwaits or 3-ways?  That's clever...?\n");
                ast_log(LOG_WARNING,
                        "p0: chan %d/%d/CW%d/3W%d, p1: chan %d/%d/CW%d/3W%d\n",
                        p0->channel,
                        oi1,
                        (p0->subs[SUB_CALLWAIT].zfd >= 0)  ?  1  :  0,
                        p0->subs[SUB_REAL].inthreeway,
                        p0->channel,
                        oi1,
                        (p1->subs[SUB_CALLWAIT].zfd >= 0)  ?  1  :  0,
                        p1->subs[SUB_REAL].inthreeway);
            }
            /*endif*/
        }
        /*endif*/
    }
    else if (oi1 == SUB_REAL  &&  oi2 == SUB_THREEWAY)
    {
        if (p1->subs[SUB_THREEWAY].inthreeway)
        {
            master = p1;
            slave = p0;
        }
        else
        {
            nothingok = 1;
        }
        /*endif*/
    }
    else if (oi1 == SUB_THREEWAY  &&  oi2 == SUB_REAL)
    {
        if (p0->subs[SUB_THREEWAY].inthreeway)
        {
            master = p0;
            slave = p1;
        }
        else
        {
            nothingok  = 1;
        }
        /*endif*/
    }
    else if (oi1 == SUB_REAL  &&  oi2 == SUB_CALLWAIT)
    {
        /* We have a real and a call wait.  If we're in a three way call, put us in it, otherwise, 
           don't put us in anything */
        if (p1->subs[SUB_CALLWAIT].inthreeway)
        {
            master = p1;
            slave = p0;
        }
        else
        {
            nothingok = 1;
        }
        /*endif*/
    }
    else if (oi1 == SUB_CALLWAIT  &&  oi2 == SUB_REAL)
    {
        /* Same as previous */
        if (p0->subs[SUB_CALLWAIT].inthreeway)
        {
            master = p0;
            slave = p1;
        }
        else
        {
            nothingok = 1;
        }
        /*endif*/
    }
    /*endif*/
    if (master  &&  slave)
    {
        /* Stop any tones, or play ringtone as appropriate.  If they're bridged
           in an active threeway call with a channel that is ringing, we should
           indicate ringing. */
        if (oi2 == SUB_THREEWAY
            && 
            p1->subs[SUB_THREEWAY].inthreeway
            && 
            p1->subs[SUB_REAL].owner
            && 
            p1->subs[SUB_REAL].inthreeway
            &&
            p1->subs[SUB_REAL].owner->_state == AST_STATE_RINGING)
        {
            ast_log(LOG_DEBUG, "Playing ringback on %s since %s is in a ringing three-way\n", c0->name, c1->name);
            tone_zone_play_tone(p0->subs[oi1].zfd, ZT_TONE_RINGTONE);
            os2 = p1->subs[SUB_REAL].owner->_state;
        }
        else
        {
            ast_log(LOG_DEBUG, "Stoping tones on %d/%d talking to %d/%d\n", p0->channel, oi1, p1->channel, oi2);
            tone_zone_play_tone(p0->subs[oi1].zfd, -1);
        }
        /*endif*/
        if (oi1 == SUB_THREEWAY
            && 
            p0->subs[SUB_THREEWAY].inthreeway
            && 
            p0->subs[SUB_REAL].owner
            && 
            p0->subs[SUB_REAL].inthreeway
            && 
            p0->subs[SUB_REAL].owner->_state == AST_STATE_RINGING)
        {
            ast_log(LOG_DEBUG, "Playing ringback on %s since %s is in a ringing three-way\n", c1->name, c0->name);
            tone_zone_play_tone(p1->subs[oi2].zfd, ZT_TONE_RINGTONE);
            os1 = p0->subs[SUB_REAL].owner->_state;
        }
        else
        {
            ast_log(LOG_DEBUG, "Stoping tones on %d/%d talking to %d/%d\n", p1->channel, oi2, p0->channel, oi1);
            tone_zone_play_tone(p1->subs[oi1].zfd, -1);
        }
        /*endif*/
        unicall_link(slave, master);
        master->inconference = inconf;
    }
    else if (!nothingok)
    {
        ast_log(LOG_WARNING, "Can't link %d/%s with %d/%s\n", p0->channel, subnames[oi1], p1->channel, subnames[oi2]);
    }
    /*endif*/

    update_conf(p0);
    update_conf(p1);
    t1 = p0->subs[SUB_REAL].inthreeway;
    t2 = p1->subs[SUB_REAL].inthreeway;

    ast_mutex_unlock(&p0->lock);
    ast_mutex_unlock(&p1->lock);

    ast_mutex_unlock(&c0->lock);
    ast_mutex_unlock(&c1->lock);

    /* Native bridge failed */
    if ((!master  ||  !slave)  &&  !nothingok)
        return -1;
    /*endif*/
    cs[SUB_REAL] = c0;
    cs[SUB_CALLWAIT] = c1;
    cs[SUB_THREEWAY] = NULL;
    for (;;)
    {
        /* Here's our main loop...  Start by locking things, looking for private parts, 
           and then balking if anything is wrong */
        ast_mutex_lock(&c0->lock);
        ast_mutex_lock(&c1->lock);
        p0 = c0->pvt->pvt;
        p1 = c1->pvt->pvt;
        if (op0 == p0)
            i1 = unicall_get_index(c0, p0, 1);
        /*endif*/
        if (op1 == p1)
            i2 = unicall_get_index(c1, p1, 1);
        /*endif*/
        ast_mutex_unlock(&c0->lock);
        ast_mutex_unlock(&c1->lock);
        if (op0 != p0
            ||
            op1 != p1
            ||
            ofd1 != c0->fds[0]
            ||
            ofd2 != c1->fds[0]
            ||
            (p0->subs[SUB_REAL].owner  &&  (os1 > -1)  &&  (os1 != p0->subs[SUB_REAL].owner->_state))
            ||
            (p1->subs[SUB_REAL].owner  &&  (os2 > -1)  &&  (os2 != p1->subs[SUB_REAL].owner->_state))
            ||
            oc1 != p0->owner
            || 
            oc2 != p1->owner
            ||
            t1 != p0->subs[SUB_REAL].inthreeway
            ||
            t2 != p1->subs[SUB_REAL].inthreeway
            ||
            oi1 != i1
            ||
            oi2 != i2)
        {
            if (slave  &&  master)
                unicall_unlink(slave, master);
            /*endif*/
            ast_log(LOG_DEBUG,
                    "Something changed out on %d/%d to %d/%d, returning -3 to restart\n",
                    op0->channel,
                    oi1,
                    op1->channel,
                    oi2);
            return -3;
        }
        /*endif*/
        to = -1;
        if ((who = ast_waitfor_n(cs, 2, &to)) == NULL)
        {
            ast_log(LOG_DEBUG, "Ooh, empty read...\n");
            continue;
        }
        /*endif*/
        if (who->pvt->pvt == op0) 
            op0->ignoredtmf = 1;
        else if (who->pvt->pvt == op1)
            op1->ignoredtmf = 1;
        /*endif*/
        f = ast_read(who);
        if (who->pvt->pvt == op0) 
            op0->ignoredtmf = 0;
        else if (who->pvt->pvt == op1)
            op1->ignoredtmf = 0;
        /*endif*/
        if (!f)
        {
            *fo = NULL;
            *rc = who;
            if (slave  &&  master)
                unicall_unlink(slave, master);
            /*endif*/
            return 0;
        }
        /*endif*/
        if (f->frametype == AST_FRAME_DTMF)
        {
            if ((who == c0  &&  (flags & AST_BRIDGE_DTMF_CHANNEL_0))
                || 
                (who == c1  &&  (flags & AST_BRIDGE_DTMF_CHANNEL_1)))
            {
                *fo = f;
                *rc = who;
                if (slave  &&  master)
                    unicall_unlink(slave, master);
                /*endif*/
                return 0;
            }
            /*endif*/
        }
        /*endif*/
        ast_frfree(f);

        /* Swap who gets priority */
        cs[2] = cs[0];
        cs[0] = cs[1];
        cs[1] = cs[2];
    }
    /*endif*/
}

static int unicall_indicate(struct ast_channel *chan, int condition);

static int unicall_fixup(struct ast_channel *oldchan, struct ast_channel *newchan)
{
    struct unicall_pvt *p = newchan->pvt->pvt;
    int x;

    ast_log(LOG_WARNING, "New owner for channel %d is %s\n", p->channel, newchan->name);
    //ast_log(LOG_DEBUG, "New owner for channel %d is %s\n", p->channel, newchan->name);
    if (p->owner == oldchan)
        p->owner = newchan;
    /*endif*/
    for (x = 0;  x < 3;  x++)
    {
        if (p->subs[x].owner == oldchan)
        {
            if (!x)
                unicall_unlink(NULL, p);
            /*endif*/
            p->subs[x].owner = newchan;
        }
        /*endif*/
    }
    /*endfor*/
    if (newchan->_state == AST_STATE_RINGING) 
        unicall_indicate(newchan, AST_CONTROL_RINGING);
    /*endif*/
    update_conf(p);
    return 0;
}

static struct ast_channel *unicall_new(struct unicall_pvt *, int, int, int, int);

struct ast_frame *unicall_exception(struct ast_channel *ast)
{
    struct unicall_pvt *p = ast->pvt->pvt;
    int index;

    if ((index = unicall_get_index(ast, p, 0)) < 0)
        return  NULL;
    /*endif*/
    
    p->subs[index].f.frametype = AST_FRAME_NULL;
    p->subs[index].f.datalen = 0;
    p->subs[index].f.samples = 0;
    p->subs[index].f.mallocd = 0;
    p->subs[index].f.offset = 0;
    p->subs[index].f.subclass = 0;
    p->subs[index].f.src = "unicall_exception";
    p->subs[index].f.data = NULL;
    
    if (p->owner == NULL  &&  !p->radio)
    {
        /* If nobody owns us, absorb the event appropriately, otherwise
           we loop indefinitely.  This occurs when, during call waiting, the
           other end hangs up our channel so that it no longer exists, but we
           have neither FLASH'd nor ONHOOK'd to signify our desire to
           change to the other channel. */
        uc_CheckEvent(p->uc);
        uc_schedule_run(p->uc);
        /* TODO: process event properly */
        return &p->subs[index].f;
    }
    /*endif*/
    if (!p->radio)
        ast_log(LOG_DEBUG, "Exception on %d, channel %d\n", ast->fds[0],p->channel);
    /*endif*/
    /* If it's not us, return NULL immediately */
    if (ast != p->owner)
    {
 ast_log(LOG_WARNING, "ast = %p, p->owner = %p\n", ast, p->owner);
        ast_log(LOG_WARNING, "We're %s, not %s\n", ast->name, p->owner->name);
        return &p->subs[index].f;
    }
    /*endif*/
    uc_CheckEvent(p->uc);
    uc_schedule_run(p->uc);
    return &p->subs[index].f;
}

void handle_uc_read(uc_t *uc, int ch, int user_data, uint8_t *buf, int len)
{
    void *readbuf;
    struct unicall_pvt *p;
    
    p = (struct unicall_pvt *) user_data;

    readbuf = ((uint8_t *) p->subs[SUB_REAL].buffer) + AST_FRIENDLY_OFFSET;
    memcpy(readbuf, buf, len);
    p->subs[SUB_REAL].buffer_contents = len;
}

struct ast_frame *unicall_read(struct ast_channel *ast)
{
    struct unicall_pvt *p = ast->pvt->pvt;
    int index;
    void *readbuf;
    struct ast_frame *f;
    char tmpfax[256];

//ast_log(LOG_WARNING, "unicall_read\n");
    ast_mutex_lock(&p->lock);
    
    index = unicall_get_index(ast, p, 0);
    
    /* Hang up if we don't really exist */
    if (index < 0)
    {
        ast_log(LOG_WARNING, "We don't exist?\n");
        pthread_mutex_unlock(&p->lock);
        return NULL;
    }
    /*endif*/
    
    p->subs[index].f.frametype = AST_FRAME_NULL;
    p->subs[index].f.datalen = 0;
    p->subs[index].f.samples = 0;
    p->subs[index].f.mallocd = 0;
    p->subs[index].f.offset = 0;
    p->subs[index].f.subclass = 0;
    p->subs[index].f.src = "unicall_read";
    p->subs[index].f.data = NULL;
    
    /* Make sure it sends initial key state as first frame */
    if (p->radio  &&  !p->firstradio)
    {
        ZT_PARAMS ps;

        ps.channo = p->channel;
        if (ioctl(p->subs[SUB_REAL].zfd, ZT_GET_PARAMS, &ps))
            return NULL;
        /*endif*/
        p->firstradio = 1;
        p->subs[index].f.frametype = AST_FRAME_CONTROL;
        if (ps.rxisoffhook)
            p->subs[index].f.subclass = AST_CONTROL_RADIO_KEY;
        else
            p->subs[index].f.subclass = AST_CONTROL_RADIO_UNKEY;
        /*endif*/
        pthread_mutex_unlock(&p->lock);
        return &p->subs[index].f;
    }
    /*endif*/
    if (p->subs[index].needringing)
    {
        /* Send ringing frame if requested */
        p->subs[index].needringing = 0;
        p->subs[index].f.frametype = AST_FRAME_CONTROL;
        p->subs[index].f.subclass = AST_CONTROL_RINGING;
        ast_setstate(ast, AST_STATE_RINGING);
        pthread_mutex_unlock(&p->lock);
        return &p->subs[index].f;
    }
    /*endif*/

    if (p->subs[index].needanswer)
    {
ast_log(LOG_WARNING, "needanswer\n");
        /* Send answer frame if requested */
        p->subs[index].needanswer = 0;
        p->subs[index].f.frametype = AST_FRAME_CONTROL;
        p->subs[index].f.subclass = AST_CONTROL_ANSWER;
        ast_setstate(ast, AST_STATE_UP);
        pthread_mutex_unlock(&p->lock);
        return &p->subs[index].f;
    }
    /*endif*/
    
    /* Check first for any outstanding DTMF characters */
    if (p->subs[index].dtmfq[0])
    {
ast_log(LOG_WARNING, "got a DTMF %d\n", p->subs[index].dtmfq[0]);
        p->subs[index].f.subclass = p->subs[index].dtmfq[0];
        memmove(&p->subs[index].dtmfq[0], &p->subs[index].dtmfq[1], sizeof(p->subs[index].dtmfq) - 1);
        p->subs[index].f.frametype = AST_FRAME_DTMF;
        if (p->subs[index].f.subclass == 'f')
        {
            /* Fax tone -- Handle and return NULL */
            if (!p->faxhandled)
            {
                p->faxhandled++;
                if (strncasecmp(ast->exten, "fax", 3))
                {
                    snprintf(tmpfax, sizeof(tmpfax), "fax%s", ast->exten);
                    if (ast_exists_extension(ast, ast->context, tmpfax, 1, ast->callerid))
                    {
                        if (option_verbose > 2)
                            ast_verbose(VERBOSE_PREFIX_3 "Redirecting %s to specific fax extension %s\n", ast->name, tmpfax);
                        /*endif*/
                        if (ast_async_goto(ast, ast->context, tmpfax, 1))
                            ast_log(LOG_WARNING, "Failed to async goto '%s' into '%s' of '%s'\n", ast->name, tmpfax, ast->context);
                        /*endif*/
                    }
                    else if (ast_exists_extension(ast, ast->context, "fax", 1, ast->callerid))
                    {
                        if (option_verbose > 2)
                            ast_verbose(VERBOSE_PREFIX_3 "Redirecting %s to fax extension\n", ast->name);
                        /*endif*/
                        if (ast_async_goto(ast, ast->context, "fax", 1))
                            ast_log(LOG_WARNING, "Failed to async goto '%s' into fax of '%s'\n", ast->name, ast->context);
                        /*endif*/
                    }
                    else
                    {
                        ast_log(LOG_NOTICE, "Fax detected, but no fax extension\n");
                    }
                    /*endif*/
                }
                else
                {
                    ast_log(LOG_DEBUG, "Already in a fax extension, not redirecting\n");
                }
                /*endif*/
            }
            else
            {
                ast_log(LOG_DEBUG, "Fax already handled\n");
            }
            /*endif*/
            p->subs[index].f.frametype = AST_FRAME_NULL;
            p->subs[index].f.subclass = 0;
            return &p->subs[index].f;
        }
        /*endif*/
        if (p->confirmanswer)
        {
            printf("Confirm answer!\n");
            /* Upon receiving a DTMF digit, consider this an answer confirmation instead
               of a DTMF digit */
            p->subs[index].f.frametype = AST_FRAME_CONTROL;
            p->subs[index].f.subclass = AST_CONTROL_ANSWER;
            ast_setstate(ast, AST_STATE_UP);
        }
        /*endif*/
        pthread_mutex_unlock(&p->lock);
        return &p->subs[index].f;
    }
    /*endif*/

    select_codec(p, index, ast->pvt->rawreadformat);

    readbuf = ((unsigned char *) p->subs[index].buffer) + AST_FRIENDLY_OFFSET;
    CHECK_BLOCKING(ast);

    uc_CheckEvent(p->uc);
    uc_schedule_run(p->uc);

    ast->blocking = 0;

    if (p->subs[index].buffer_contents == 0)
    {
        p->subs[index].f.frametype = AST_FRAME_NULL;
        p->subs[index].f.subclass = 0;
        p->subs[index].f.samples = 0;
        p->subs[index].f.mallocd = 0;
        p->subs[index].f.offset = 0;
        p->subs[index].f.data = NULL;
        p->subs[index].f.datalen= 0;
        pthread_mutex_unlock(&p->lock);
        return &p->subs[index].f;
    }
    /*endif*/

    if (ast->pvt->rawreadformat == AST_FORMAT_SLINEAR)
        p->subs[index].f.datalen = READ_SIZE*2;
    else 
        p->subs[index].f.datalen = READ_SIZE;
    /*endif*/
    p->subs[index].f.frametype = AST_FRAME_VOICE;
    p->subs[index].f.subclass = ast->pvt->rawreadformat;
    p->subs[index].f.samples = READ_SIZE;
    p->subs[index].f.mallocd = 0;
    p->subs[index].f.offset = AST_FRIENDLY_OFFSET;
    p->subs[index].f.data = p->subs[index].buffer + AST_FRIENDLY_OFFSET/2;
#if 0
    ast_log(LOG_DEBUG, "Read %d of voice on %s\n", p->subs[index].f.datalen, ast->name);
#endif
    if (p->dialing
        ||
        (index  &&  ast->_state != AST_STATE_UP)
        ||
        (index == SUB_CALLWAIT  &&  !p->subs[SUB_CALLWAIT].inthreeway))
    {
        /* Whoops, we're still dialing, or in a state where we shouldn't transmit....
           don't send anything */
        p->subs[index].f.frametype = AST_FRAME_NULL;
        p->subs[index].f.subclass = 0;
        p->subs[index].f.samples = 0;
        p->subs[index].f.mallocd = 0;
        p->subs[index].f.offset = 0;
        p->subs[index].f.data = NULL;
        p->subs[index].f.datalen= 0;
    }
    /*endif*/
    if (p->dsp  &&  !p->ignoredtmf  &&  index == SUB_REAL)
    {
        if ((f = ast_dsp_process(ast, p->dsp, &p->subs[index].f)))
        {
            if ((f->frametype == AST_FRAME_CONTROL)  &&  (f->subclass == AST_CONTROL_BUSY))
            {
                if ((ast->_state == AST_STATE_UP)  &&  !p->outgoing)
                {
                    /* Treat this as a "hangup" instead of a "busy" on the assumption that
		       its a busy  */
                    f = NULL;
                }
                /*endif*/
            }
            /*endif*/
        }
        /*endif*/
    }
    else 
    {
        f = &p->subs[index].f; 
    }
    /*endif*/
    if (f  &&  (f->frametype == AST_FRAME_DTMF))
    {
        ast_log(LOG_DEBUG, "DTMF digit: %c on %s\n", f->subclass, ast->name);
        if (p->confirmanswer)
        {
            ast_log(LOG_DEBUG, "Confirm answer on %s!\n", ast->name);
            /* Upon receiving a DTMF digit, consider this an answer confirmation instead
               of a DTMF digit */
            p->subs[index].f.frametype = AST_FRAME_CONTROL;
            p->subs[index].f.subclass = AST_CONTROL_ANSWER;
            ast_setstate(ast, AST_STATE_UP);
            f = &p->subs[index].f;
        }
        else if (f->subclass == 'f')
        {
            /* Fax tone -- Handle and return NULL */
            if (!p->faxhandled)
            {
		p->faxhandled++;
		if (strcmp(ast->exten, "fax"))
                {
		    if (ast_exists_extension(ast, ast->context, "fax", 1, ast->callerid))
                    {
			if (option_verbose > 2)
			    ast_verbose(VERBOSE_PREFIX_3 "Redirecting %s to fax extension\n", ast->name);
			/*endif*/
                        if (ast_async_goto(ast, ast->context, "fax", 1))
			    ast_log(LOG_WARNING, "Failed to async goto '%s' into fax of '%s'\n", ast->name, ast->context);
                        /*endif*/
		    }
                    else
                    {
			ast_log(LOG_NOTICE, "Fax detected, but no fax extension\n");
                    }
                    /*endif*/
		}
                else
                {
		    ast_log(LOG_DEBUG, "Already in a fax extension, not redirecting\n");
                }
                /*endif*/
	    }
            else
            {
                ast_log(LOG_DEBUG, "Fax already handled\n");
            }
            /*endif*/
	    unicall_confmute(p, 0);
	    p->subs[index].f.frametype = AST_FRAME_NULL;
	    p->subs[index].f.subclass = 0;
	    f = &p->subs[index].f;
	}
        else if (f->subclass == 'm')
        {
	    /* Confmute request */
	    unicall_confmute(p, 1);
	    p->subs[index].f.frametype = AST_FRAME_NULL;
	    p->subs[index].f.subclass = 0;
	    f = &p->subs[index].f;		
	}
        else if (f->subclass == 'u')
        {
	    /* Unmute */
	    unicall_confmute(p, 0);
	    p->subs[index].f.frametype = AST_FRAME_NULL;
	    p->subs[index].f.subclass = 0;
	    f = &p->subs[index].f;		
	}
        else
        {
	    unicall_confmute(p, 0);
        }
        /*endif*/
    }
    /*endif*/

    pthread_mutex_unlock(&p->lock);
    return f;
}

static int unicall_write(struct ast_channel *ast, struct ast_frame *frame)
{
    struct unicall_pvt *p = ast->pvt->pvt;
    int index;
    int len;
    int size;
    int res;
    unsigned char *buf;

    res = 0;    
//ast_log(LOG_WARNING, "unicall_write\n");
    if ((index = unicall_get_index(ast, p, 0)) < 0)
    {
        ast_log(LOG_WARNING, "%s doesn't really exist?\n", ast->name);
        return -1;
    }
    /*endif*/    
    /* Write a frame of (presumably voice) data */
    if (frame->frametype != AST_FRAME_VOICE)
    {
        if (frame->frametype != AST_FRAME_IMAGE)
            ast_log(LOG_WARNING, "Don't know what to do with frame type '%d'\n", frame->frametype);
        /*endif*/
        return 0;
    }
    /*endif*/
    if ((frame->subclass != AST_FORMAT_SLINEAR)
        && 
        (frame->subclass != AST_FORMAT_ULAW)
        &&
        (frame->subclass != AST_FORMAT_ALAW))
    {
        ast_log(LOG_WARNING, "Cannot handle frames in %d format\n", frame->subclass);
        return -1;
    }
    /*endif*/
    if (p->dialing)
    {
        if (option_debug)
            ast_log(LOG_DEBUG, "Dropping frame since I'm still dialing on %s...\n", ast->name);
        /*endif*/
        return 0;
    }
    /*endif*/
    /* Return if it's not valid data */
    if (frame->data == NULL  ||  frame->datalen == 0)
        return 0;
    /*endif*/

    select_codec(p, index, frame->subclass);

    len = frame->datalen;
    buf = (unsigned char *) frame->data;
    while (len)
    {
        size = len;
        if (size > READ_SIZE)
            size = READ_SIZE;
        /*endif*/
        if ((res = write(p->subs[index].zfd, buf, size)) != size)
        {
            if (res < 0  &&  errno == ELAST)
            {
                uc_CheckEvent(p->uc);
                uc_schedule_run(p->uc);
                continue;
            }
            /*endif*/
            if (option_debug)
                ast_log(LOG_DEBUG, "Write returned %d (%s) on channel %d\n", res, strerror(errno), p->channel);
            /*endif*/
            break;
        }
        /*endif*/
        len -= res;
        buf += res;
    }
    /*endwhile*/
    if (res < 0  &&  errno != EAGAIN)
    {
        ast_log(LOG_WARNING, "write failed: %s - %d\n", strerror(errno), errno);
        return -1;
    }
    /*endif*/
    return 0;
}

static int unicall_indicate(struct ast_channel *chan, int condition)
{
    struct unicall_pvt *p = chan->pvt->pvt;
    int res = -1;
    int index;

    if ((index = unicall_get_index(chan, p, 0)) != SUB_REAL)
        return 0;
    /*endif*/
    switch (condition)
    {
    case AST_CONTROL_BUSY:
        res = tone_zone_play_tone(p->subs[index].zfd, ZT_TONE_BUSY);
        break;
    case AST_CONTROL_RINGING:
        res = tone_zone_play_tone(p->subs[index].zfd, ZT_TONE_RINGTONE);
        if (chan->_state != AST_STATE_UP)
        {
            if (chan->_state != AST_STATE_RING)
                ast_setstate(chan, AST_STATE_RINGING);
            /*endif*/
        }
        /*endif*/
        break;
    case AST_CONTROL_CONGESTION:
        res = tone_zone_play_tone(p->subs[index].zfd, ZT_TONE_CONGESTION);
        break;
    case AST_CONTROL_RADIO_KEY:
        if (p->radio)
        {
            if (uc_DropCall(p->uc, p->crn, UC_CAUSE_NORMAL_CLEARING) < 0)
		ast_log(LOG_WARNING, "uc_DropCall failed\n");
            /*endif*/
        }
        /*endif*/
        res = 0;
        break;
    case AST_CONTROL_RADIO_UNKEY:
        if (p->radio)
        {
            if (uc_DropCall(p->uc, p->crn, UC_CAUSE_NORMAL_CLEARING) < 0)
		ast_log(LOG_WARNING, "uc_DropCall failed\n");
            /*endif*/
        }
        /*endif*/
        res = 0;
        break;
    case -1:
        res = tone_zone_play_tone(p->subs[index].zfd, -1);
        break;
    default:
        ast_log(LOG_WARNING, "Don't know how to set condition %d on channel %s\n", condition, chan->name);
        break;
    }
    /*endswitch*/
    return res;
}

static struct ast_channel *unicall_new(struct unicall_pvt *i, int state, int startpbx, int index, int law)
{
    struct ast_channel *tmp;
    int deflaw;
    int res;
    int x;
    int y;
    ZT_PARAMS ps;

    if ((tmp = ast_channel_alloc(0)) == NULL)
    {
        ast_log(LOG_WARNING, "Unable to allocate channel structure\n");
        return  NULL;
    }
    /*endif*/
    ps.channo = i->channel;
    if ((res = ioctl(i->subs[SUB_REAL].zfd, ZT_GET_PARAMS, &ps)))
    {
        ast_log(LOG_WARNING, "Unable to get parameters, assuming MULAW\n");
        ps.curlaw = ZT_LAW_MULAW;
    }
    /*endif*/

    if (law)
    {
        if (law == UC_CODEC_ALAW)
            deflaw = AST_FORMAT_ALAW;
        else
            deflaw = AST_FORMAT_ULAW;
        /*endif*/
    }
    else
    {
        if (ps.curlaw == ZT_LAW_ALAW)
            deflaw = AST_FORMAT_ALAW;
        else
            deflaw = AST_FORMAT_ULAW;
        /*endif*/
    }
    /*endif*/
    y = 1;
    do
    {
        snprintf(tmp->name, sizeof(tmp->name), "UniCall/%d-%d", i->channel, y);
        for (x = 0;  x < 3;  x++)
        {
            if (index != x  &&  i->subs[x].owner  &&  strcasecmp(tmp->name, i->subs[x].owner->name) == 0)
                break;
            /*endif*/
        }
        /*endfor*/
        y++;
    }
    while (x < 3);
    tmp->type = type;
    tmp->fds[0] = i->subs[index].zfd;
    tmp->nativeformats = AST_FORMAT_SLINEAR | deflaw;
    /* Start out assuming ulaw since it's smaller :) */
    tmp->pvt->rawreadformat = deflaw;
    tmp->readformat = deflaw;
    tmp->pvt->rawwriteformat = deflaw;
    tmp->writeformat = deflaw;
    i->subs[index].current_codec = deflaw;
    if (uc_set_api_codec(i->uc, 0, UC_CODEC_DEFAULT))
        ast_log(LOG_WARNING, "Unable to set law on channel %d to default\n", i->channel);
    /*endif*/

    if (i->dsp)
    {
        ast_log(LOG_DEBUG, "Already have a dsp on %s?\n", tmp->name);
    }
    else
    {
        if ((i->dsp = ast_dsp_new()))
        {
            ast_dsp_set_features(i->dsp, DSP_FEATURE_DTMF_DETECT);
            ast_dsp_digitmode(i->dsp, DSP_DIGITMODE_DTMF | i->dtmfrelax);
        }
        /*endif*/
    }
    /*endif*/

    if (state == AST_STATE_RING)
        tmp->rings = 1;
    /*endif*/
    tmp->pvt->pvt = i;
    tmp->pvt->send_digit = unicall_digit;
    tmp->pvt->send_text = unicall_sendtext;
    tmp->pvt->call = unicall_call;
    tmp->pvt->hangup = unicall_hangup;
    tmp->pvt->answer = unicall_answer;
    tmp->pvt->read = unicall_read;
    tmp->pvt->write = unicall_write;
    tmp->pvt->bridge = unicall_bridge;
    tmp->pvt->exception = unicall_exception;
    tmp->pvt->indicate = unicall_indicate;
    tmp->pvt->fixup = unicall_fixup;
    tmp->pvt->setoption = unicall_setoption;
    if (strlen(i->language))
        strncpy(tmp->language, i->language, sizeof(tmp->language) - 1);
    /*endif*/
    if (strlen(i->musicclass))
        strncpy(tmp->musicclass, i->musicclass, sizeof(tmp->musicclass) - 1);
    /*endif*/
    if (i->owner == NULL)
        i->owner = tmp;
    /*endif*/
    if (strlen(i->accountcode))
        strncpy(tmp->accountcode, i->accountcode, sizeof(tmp->accountcode) - 1);
    /*endif*/
    if (i->amaflags)
        tmp->amaflags = i->amaflags;
    /*endif*/
    if (i->subs[index].owner)
        ast_log(LOG_WARNING, "Channel %d already has a %s call\n", i->channel,subnames[index]);
    /*endif*/
    i->subs[index].owner = tmp;
    ast_setstate(tmp, state);
    ast_mutex_lock(&usecnt_lock);
    usecnt++;
    ast_mutex_unlock(&usecnt_lock);
    ast_update_use_count();
    strncpy(tmp->context, i->context, sizeof(tmp->context) - 1);
    /* Copy call forward info */
    strncpy(tmp->call_forward, i->call_forward, sizeof(tmp->call_forward));
    /* If we've been told "no ADSI" then enforce it */
    if (!i->adsi)
        tmp->adsicpe = AST_ADSI_UNAVAILABLE;
    /*endif*/
    if (strlen(i->exten))
        strncpy(tmp->exten, i->exten, sizeof(tmp->exten) - 1);
    /*endif*/
    if (strlen(i->callerid))
    {
        tmp->callerid = strdup(i->callerid);
        tmp->ani = strdup(i->callerid);
    }
    /*endif*/
    if (strlen(i->rdnis))
        tmp->rdnis = strdup(i->rdnis);
    /*endif*/
    if (startpbx  &&  ast_pbx_start(tmp))
    {
        ast_log(LOG_WARNING, "Unable to start PBX on %s\n", tmp->name);
        ast_hangup(tmp);
        tmp = NULL;
    }
    /*endif*/
    return tmp;
}

int channel_error(uc_t *uc, int user_data, int cause)
{
    ast_log(LOG_ERROR, "Channel error - %d\n", cause);
    return  0;
}

void handle_uc_event(uc_t *uc, int user_data, uc_event_t *ev)
{
    struct ast_channel *c;
    struct unicall_pvt *i;
    int ch;
    
    ch = 0;
    i = (struct unicall_pvt *) user_data;
    if (i->radio)
        return;
    /*endif*/
    /* Handle an event on a given channel for the monitor thread. */

    ast_log(LOG_WARNING, "UC event %s\n", uc_event2str(ev->e));
    switch (ev->e)
    {
    case UC_EVENT_PROTOCOLFAIL:
        if (option_verbose > 2)
            ast_verbose(VERBOSE_PREFIX_3 "UC channel %d protocol error. Cause %d\n", i->channel, ev->gen.data);
        /*endif*/
        break;
    case UC_EVENT_LOCALUNBLOCKED:
        if (option_verbose > 2)
            ast_verbose(VERBOSE_PREFIX_3 "UC channel %d local unblocked\n", i->channel);
        /*endif*/
        break;
    case UC_EVENT_LOCALBLOCKED:
        if (option_verbose > 2)
            ast_verbose(VERBOSE_PREFIX_3 "UC channel %d local blocked\n", i->channel);
        /*endif*/
        break;
    case UC_EVENT_FARUNBLOCKED:
        i->blocked = 0;
        if (option_verbose > 2)
            ast_verbose(VERBOSE_PREFIX_3 "UC channel %d far unblocked\n", i->channel);
        /*endif*/
        break;
    case UC_EVENT_FARBLOCKED:
        i->blocked = 1;
        if (option_verbose > 2)
            ast_verbose(VERBOSE_PREFIX_3 "UC channel %d far unblocked\n", i->channel);
        /*endif*/
        break;
    case UC_EVENT_DETECTED:
        /* A call is beginning to come in */
        i->reserved = 1;
        break;
    case UC_EVENT_DIALTONE:
        break;
    case UC_EVENT_DIALING:
        break;
    case UC_EVENT_PROCEEDING:
        break;
    case UC_EVENT_OFFERED:
        /* Check for callerid, digits, etc */
ast_log(LOG_WARNING, "Offered on channel %d (ANI: %s, DNIS: %s)\n", ev->offered.channel, ev->offered.parms.originating_number, ev->offered.parms.destination_number);
#if 0
        if ((ch = ev->offered.channel))
        {
#endif
            /* Get caller ID */
            if (i->use_callerid)
            {
                if (ev->offered.parms.originating_name[0])
                    snprintf(i->callerid, sizeof(i->callerid), "\"%s\" <%s>", ev->offered.parms.originating_name, ev->offered.parms.originating_number);
                else
                    strncpy(i->callerid, ev->offered.parms.originating_number, sizeof(i->callerid) - 1);
                /*endif*/
            }
            else
            {
                strcpy(i->callerid, "");
            }
            /*endif*/
            i->callerid[sizeof(i->callerid) - 1] = '\0';
            if (i->owner)
            {
                i->owner->callerid = strdup(i->callerid);
                i->owner->ani = strdup(i->callerid);
            }
            /*endif*/
            strncpy(i->rdnis, ev->offered.parms.redirecting_number, sizeof(i->rdnis));
            i->rdnis[sizeof(i->rdnis) - 1] = '\0';
            /* Get called number */
            if (strlen(ev->offered.parms.destination_number))
                strncpy(i->exten, ev->offered.parms.destination_number, sizeof(i->exten) - 1);
            else
                strcpy(i->exten, "s");
            /*endif*/
            /* Make sure extension exists */
            if (ast_exists_extension(NULL, i->context, i->exten, 1, i->callerid))
            {
                //i->call = ev->offered.call;
                i->crn = ev->offered.crn;
                /* Setup law */
                if (ev->offered.parms.userinfo_layer1_protocol == UC_LAYER_1_ALAW)
                    i->law = UC_CODEC_ALAW;
                else
                    i->law = UC_CODEC_ULAW;
                /*endif*/
#if 1
i->law = UC_CODEC_ALAW;
#endif
                uc_AcceptCall(i->uc, ev->offered.crn);
            }
            else
            {
                if (ast_matchmore_extension(NULL, i->context, i->exten, 1, i->callerid))
                {
                    if (ev->e == UC_EVENT_OFFERED)
                        uc_CallAck(i->uc, ev->offered.crn, NULL);
                    /*endif*/
                }
                else
                {
                    if (option_verbose > 2)
                    {
                        ast_verbose(VERBOSE_PREFIX_3 "Extension '%s' in context '%s' from '%s' does not exist.  Rejecting call on channel %d\n",
                                    i->exten,
                                    i->context,
                                    i->callerid,
                                    i->channel);
                    }
                    /*endif*/
                    if (uc_DropCall(i->uc, ev->offered.crn, UC_CAUSE_UNASSIGNED_NUMBER) < 0)
                        ast_log(LOG_WARNING, "uc_DropCall failed\n");
                    /*endif*/
                }
                /*endif*/
            }
            /*endif*/
#if 0
        }
        else
        { 
            if (uc_DropCall(i->uc, ev->offered.call, UC_CAUSE_CHANNEL_UNACCEPTABLE) < 0)
                ast_log(LOG_WARNING, "uc_DropCall failed\n");
            /*endif*/
        }
        /*endif*/
#endif
        break;
    case UC_EVENT_MOREDIGITS:
        ast_log(LOG_WARNING, "Dest '%s'\n", ev->offered.parms.destination_number);
#if 0
        if ((ch = ev->offered.channel))
        {
#endif
            /* Get caller ID */
            if (i->use_callerid)
            {
                if (strlen(ev->offered.parms.originating_name))
                    snprintf(i->callerid, sizeof(i->callerid), "\"%s\" <%s>", ev->offered.parms.originating_name, ev->offered.parms.originating_number);
                else
                    strncpy(i->callerid, ev->offered.parms.originating_number, sizeof(i->callerid) - 1);
                /*endif*/
            }
            else
            {
                strcpy(i->callerid, "");
            }
            /*endif*/
            i->callerid[sizeof(i->callerid) - 1] = '\0';
            strncpy(i->rdnis, ev->offered.parms.redirecting_number, sizeof(i->rdnis));
            i->rdnis[sizeof(i->rdnis) - 1] = '\0';
            /* Get called number */
            if (strlen(ev->offered.parms.destination_number))
                strncpy(i->exten, ev->offered.parms.destination_number, sizeof(i->exten) - 1);
            else
                strcpy(i->exten, "s");
            /*endif*/
            i->exten[sizeof(i->exten) - 1] = '\0';
            /* Make sure extension exists */
            if (ast_exists_extension(NULL, i->context, i->exten, 1, i->callerid))
            {
                //i->call = ev->offered.call;
                i->crn = ev->offered.crn;
                /* Setup law */
                if (ev->offered.parms.userinfo_layer1_protocol == UC_LAYER_1_ALAW)
                    i->law = UC_CODEC_ALAW;
                else
                    i->law = UC_CODEC_ULAW;
                /*endif*/
#if 1
i->law = UC_CODEC_ALAW;
#endif
                uc_AcceptCall(i->uc, ev->offered.crn);
            }
            else
            {
                if (ast_matchmore_extension(NULL, i->context, i->exten, 1, i->callerid))
                {
                    if (ev->e == UC_EVENT_OFFERED)
                        uc_CallAck(i->uc, ev->offered.crn, NULL);
                    /*endif*/
                }
                else
                {
                    if (option_verbose > 2)
                    {
                        ast_verbose(VERBOSE_PREFIX_3 "Extension '%s' in context '%s' from '%s' does not exist.  Rejecting call on channel %d\n",
                                    i->exten,
                                    i->context,
                                    i->callerid,
                                    i->channel);
                    }
                    /*endif*/
                    if (uc_DropCall(i->uc, ev->offered.crn, UC_CAUSE_UNASSIGNED_NUMBER) < 0)
                        ast_log(LOG_WARNING, "uc_DropCall failed\n");
                    /*endif*/
                }
                /*endif*/
            }
            /*endif*/
#if 0
        }
        else
        { 
            if (uc_DropCall(i->uc, ev->offered.call, UC_CAUSE_CHANNEL_UNACCEPTABLE) < 0)
                ast_log(LOG_WARNING, "uc_DropCall failed\n");
            /*endif*/
        }
        /*endif*/
#endif
        break;
    case UC_EVENT_ACCEPTED:        
        if (uc_set_api_codec(i->uc, 0, UC_CODEC_DEFAULT))
            ast_log(LOG_WARNING, "Unable to set law on channel %d\n", i->channel);
        /*endif*/
        if (set_actual_gain(i->subs[SUB_REAL].zfd, 0, i->rxgain, i->txgain, i->law) < 0)
            ast_log(LOG_WARNING, "Unable to set gains on channel %d\n", i->channel);
        /*endif*/
#if 1
        /* Start PBX */
        if ((c = unicall_new(i, AST_STATE_RING, 1, SUB_REAL, i->law)) == NULL)
        {
            ast_log(LOG_WARNING, "Unable to start PBX on channel %d\n", i->channel);
            if (uc_DropCall(i->uc, ev->offered.crn, UC_CAUSE_NETWORK_CONGESTION) < 0)
                ast_log(LOG_WARNING, "uc_DropCall failed\n");
            /*endif*/
        }
        /*endif*/
        i->subs[SUB_REAL].needanswer = 0;
#endif
        break;
    case UC_EVENT_ALERTING:
        ch = ev->gen.channel;
        if (ch)
        {
            if (!strlen(i->dop.dialstr))
            {
                unicall_enable_ec(i);
                i->subs[SUB_REAL].needringing = 1;
            }
            else
            {
                ast_log(LOG_DEBUG, "Deferring ringing notification because of extra digits to dial...\n");
            }
            /*endif*/
        }
        /*endif*/
        break;
    case UC_EVENT_CONNECTED:
#if 0
        if ((ch = ev->connected.channel))
        {
#endif
            if (strlen(i->dop.dialstr))
            {
                i->dialing = 1;
                /* Send any "w" waited stuff */
                if (ioctl(i->subs[SUB_REAL].zfd, ZT_DIAL, &i->dop) < 0)
                {
                    ast_log(LOG_WARNING, "Unable to initiate dialing on trunk channel %d\n", i->channel);
                    i->dop.dialstr[0] = '\0';
                    return;
                }
                /*endif*/
                ast_log(LOG_DEBUG, "Sent deferred digit string: %s\n", i->dop.dialstr);
                i->dop.dialstr[0] = '\0';
            }
            else
            {
                i->subs[SUB_REAL].needanswer = 1;
            }
            /*endif*/
            /* Enable echo cancellation if it's not on already */
            unicall_enable_ec(i);
            i->sigcheck = 0;
#if 0
        }
        else
        {
            ast_log(LOG_WARNING, "Connected/answered on bad channel %d\n", ev->connected.channel);
        }
        /*endif*/
#endif
        break;				
    case UC_EVENT_ANSWERED:
#if 0
        if ((ch = ev->connected.channel))
        {
#endif
            /* Enable echo cancellation if it's not on already */
            unicall_enable_ec(i);
#if 0
        }
        else
        {
            ast_log(LOG_WARNING, "Connected/answered on bad channel %d\n", ev->connected.channel);
        }
        /*endif*/
#endif
        break;
    case UC_EVENT_FARDISCONNECTED:
#if 0
        if ((ch = ev->fardisconnected.channel))
        {
#endif
            ast_log(LOG_WARNING, "Far disconnect cause %d\n", ev->fardisconnected.cause);
            if (i->owner)
            {
                i->alreadyhungup = 1;
                i->owner->_softhangup |= AST_SOFTHANGUP_DEV;
                if (option_verbose > 2) 
                    ast_verbose(VERBOSE_PREFIX_3 "Channel %d got hangup\n", ch);
                /*endif*/
            }
            else
            {
                if (uc_DropCall(i->uc, ev->offered.crn, UC_CAUSE_UNASSIGNED_NUMBER) < 0)
                    ast_log(LOG_WARNING, "uc_DropCall failed\n");
                /*endif*/
            }
            /*endif*/
            if (ev->fardisconnected.cause == UC_CAUSE_REQ_CHANNEL_NOT_AVAILABLE)
            {
                if (option_verbose > 2)
                    ast_verbose(VERBOSE_PREFIX_3 "Forcing restart of channel %d since channel reported in use\n", ch);
                //pri_reset(i, ch);
                //i->resetting = 1;
            }
            /*endif*/
#if 0
	}
        else
        {
            ast_log(LOG_WARNING, "Hangup on bad channel %d\n", ev->fardisconnected.channel);
        }
        /*endif*/
#endif
        break;
    case UC_EVENT_DROPCALL:
#if 0
        if ((ch = ev->gen.channel))
        {
#endif
ast_log(LOG_WARNING, "Doing a uc_ReleaseCall\n");
            if (uc_ReleaseCall(i->uc, ev->gen.crn) < 0)
                ast_log(LOG_WARNING, "uc_ReleaseCall failed\n");
            /*endif*/
#if 0
	}
        else
        {
            ast_log(LOG_WARNING, "Hangup on bad channel %d\n", ev->fardisconnected.channel);
        }
        /*endif*/
#endif
        break;
    case UC_EVENT_RELEASECALL:
#if 0
        if ((ch = ev->gen.channel))
        {
#endif
            if (option_verbose > 2)
                ast_verbose(VERBOSE_PREFIX_3 "UC channel %d released\n", i->channel);
            /*endif*/
            i->crn = 0;
            unicall_disable_ec(i);
#if 0
	}
        else
        {
            ast_log(LOG_WARNING, "Hangup on bad channel %d\n", ev->fardisconnected.channel);
        }
        /*endif*/
#endif
        i->reserved = 0;
        break;
    case UC_EVENT_ALARM:
        break;
    default:
        ast_log(LOG_WARNING,
                "Don't know how to handle signalling event %s on channel %d\n",
                uc_event2str(ev->e),
                i->channel);
        break;
    }
    /*endswitch*/
}

static void *do_monitor(void *data)
{
    struct unicall_pvt *i;
    struct unicall_pvt *last = NULL;
    fd_set efds;
    fd_set rfds;
    int n;
    int res;
    struct timeval tv;
    time_t thispass = 0;
    time_t lastpass = 0;
    int found;

    /* This thread monitors all the channels which are idle. Basically, it is waiting
       for calls to arrive. Once they do, a call thread will be created, and that will do
       all the I/O for the channel until the call ends. Then things return to this thread
       to await another call. If the channel is grabbed to make an outgoing call, this will
       also transfer I/O work to the call's thread, until the channel returns to the idle
       state. */

    /* From here on out, we die whenever asked */
#if 0
    if (pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL))
    {
        ast_log(LOG_WARNING, "Unable to set cancel type to asynchronous\n");
        return NULL;
    }
    /*endif*/
    ast_log(LOG_DEBUG, "Monitor starting...\n");
#endif
    for (;;)
    {
        /* Lock the interface list */
        if (ast_mutex_lock(&iflock))
        {
            ast_log(LOG_ERROR, "Unable to grab interface lock\n");
            return NULL;
        }
        /*endif*/
        /* Build the stuff we're going to select on. This is the socket of every
           unicall_pvt that does not have an associated owner channel. */
        n = -1;
        FD_ZERO(&efds);
        FD_ZERO(&rfds);
        for (i = iflist;  i;  i = i->next)
        {
            if (i->subs[SUB_REAL].zfd >= 0  &&  i->protocol_class  &&  !i->radio)
            {
                if (FD_ISSET(i->subs[SUB_REAL].zfd, &efds))
                    ast_log(LOG_WARNING, "Descriptor %d appears twice?\n", i->subs[SUB_REAL].zfd);
                /*endif*/
                if (i->sigcheck  ||  (i->owner == NULL  &&  i->subs[SUB_REAL].owner == NULL))
                {
                    /* This needs to be watched, as it lacks an owner */
                    FD_SET(i->subs[SUB_REAL].zfd, &rfds);
                    FD_SET(i->subs[SUB_REAL].zfd, &efds);
                    if (i->subs[SUB_REAL].zfd > n)
                        n = i->subs[SUB_REAL].zfd;
                    /*endif*/
                }
                /*endif*/
            }
            /*endif*/
        }
        /*endfor*/
        /* Okay, now that we know what to do, release the interface lock */
        ast_mutex_unlock(&iflock);
        
        pthread_testcancel();
        /* Wait at least a second for something to happen */
        tv.tv_sec = 1;
        tv.tv_usec = 0;
        res = select(n + 1, &rfds, NULL, &efds, &tv);
        pthread_testcancel();
        /* Okay, select has finished.  Let's see what happened.  */
        if (res < 0)
        {
            if (errno != EAGAIN  &&  errno != EINTR)
                ast_log(LOG_WARNING, "select return %d: %s\n", res, strerror(errno));
            /*endif*/
            continue;
        }
        /*endif*/
        /* Alright, lock the interface list again, and let's look and see what has
           happened */
        if (ast_mutex_lock(&iflock))
        {
            ast_log(LOG_WARNING, "Unable to lock the interface list\n");
            continue;
        }
        /*endif*/
        found = 0;
        lastpass = thispass;
        thispass = time(NULL);
        for (i = iflist;  i;  i = i->next)
        {
            if (thispass != lastpass)
            {
                if (!found  &&  ((i == last)  ||  ((i == iflist)  &&  !last)))
                {
                    last = i;
                    if (last)
                    {
#if 0
                        printf("Checking channel %d\n", last->channel);
#endif                        
                        if (last->owner == NULL
                            &&
                            strlen(last->mailbox)
                            &&
                            (thispass - last->onhooktime > 3)
                            &&
                            0) //(last->protocol_class & __ZT_SIG_FXO))
                        {
#if 0
                            printf("Channel %d has mailbox %s\n", last->channel, last->mailbox);
#endif                            
                            if ((res = ast_app_has_voicemail(last->mailbox)) != last->msgstate)
                            {
                                ast_log(LOG_DEBUG, "Message status for %s changed from %d to %d on %d\n", last->mailbox, last->msgstate, res, last->channel);
                                found++;
                            }
                            /*endif*/
                        }
                        /*endif*/
                        last = last->next;
                    }
                    /*endif*/
                }
                /*endif*/
            }
            /*endif*/
            if (i->subs[SUB_REAL].zfd >= 0  &&  i->protocol_class  &&  !i->radio)
            {
                if (FD_ISSET(i->subs[SUB_REAL].zfd, &rfds))
                {
                    if (!i->sigcheck  &&  (i->owner  ||  i->subs[SUB_REAL].owner))
                    {
                        ast_log(LOG_WARNING, "Whoa....  I'm owned but found (%d) in read...\n", i->subs[SUB_REAL].zfd);
                        continue;
                    }
                    /*endif*/
                    uc_CheckEvent(i->uc);
                    uc_schedule_run(i->uc);
                    continue;
                }
                /*endif*/
                if (FD_ISSET(i->subs[SUB_REAL].zfd, &efds))
                {
                    if (!i->sigcheck  &&  (i->owner  ||  i->subs[SUB_REAL].owner))
                    {
                        ast_log(LOG_WARNING, "Whoa....  I'm owned but found (%d)...\n", i->subs[SUB_REAL].zfd);
                        continue;
                    }
                    /*endif*/
                    uc_CheckEvent(i->uc);
                    uc_schedule_run(i->uc);
                }
                /*endif*/
            }
            /*endif*/
        }
        /*endfor*/
        ast_mutex_unlock(&iflock);
    }
    /*endfor*/
    /* Never reached */
    return NULL;
}

static int restart_monitor(void)
{
    pthread_attr_t attr;

    /* If we're supposed to be stopped -- stay stopped */
    if (monitor_thread == -2)
        return 0;
    /*endif*/
    if (ast_mutex_lock(&monlock))
    {
        ast_log(LOG_WARNING, "Unable to lock monitor\n");
        return -1;
    }
    /*endif*/
    if (monitor_thread == pthread_self())
    {
        ast_mutex_unlock(&monlock);
        ast_log(LOG_WARNING, "Cannot kill myself\n");
        return -1;
    }
    /*endif*/
    if (monitor_thread)
    {
        /* Just signal it to be sure it wakes up */
#if 0
        pthread_cancel(monitor_thread);
#endif
        pthread_kill(monitor_thread, SIGURG);
#if 0
        pthread_join(monitor_thread, NULL);
#endif
    }
    else
    {
ast_log(LOG_WARNING, "XXX creating monitor thread\n");
        /* Start a new monitor */
        pthread_attr_init(&attr);
        pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
        if (ast_pthread_create(&monitor_thread, &attr, do_monitor, NULL) < 0)
        {
            ast_mutex_unlock(&monlock);
            ast_log(LOG_ERROR, "Unable to start monitor thread.\n");
            return -1;
        }
        /*endif*/
    }
    /*endif*/
#if 0
    printf("Created thread %ld detached in restart monitor\n", monitor_thread);
#endif
    ast_mutex_unlock(&monlock);
    return 0;
}

static int reset_channel(struct unicall_pvt *p)
{
    int drop = 1;
    int res = 0;
    int i = 0;

    ast_log(LOG_DEBUG, "reset_channel()\n");
    if (p->owner)
    {
        drop = 0;
        p->owner->_softhangup |= AST_SOFTHANGUP_DEV;
    }
    /*endif*/
    for (i = 0;  i < 3;  i++)
    {
        if (p->subs[i].owner)
        {
            drop = 0;
            p->subs[i].owner->_softhangup |= AST_SOFTHANGUP_DEV;
        }
        /*endif*/
    }
    /*endfor*/
    if (drop)
    {
ast_log(LOG_WARNING, "DDDDD 1\n");
        if ((res = uc_DropCall(p->uc, p->crn, UC_CAUSE_NORMAL_CLEARING)))
        {
            ast_log(LOG_ERROR, "Unable to hangup chan_unicall channel %d\n", p->channel);
            return -1;
        }
        /*endif*/
    }
    /*endif*/

    return 0;
}

static struct unicall_pvt *mkintf(int channel, char *protocol_class, char *protocol_variant, char *protocol_end, int radio)
{
    /* Make a unicall_pvt structure for this interface */
    struct unicall_pvt *tmp = NULL;
    struct unicall_pvt *tmp2;
    struct unicall_pvt *prev = NULL;
    char fn[80];
    struct zt_bufferinfo bi;
    struct zt_spaninfo si;
    int res;
    int here;
    int x;
    ZT_PARAMS p;

    memset(&si, 0, sizeof(si));

    here = 0;
    for (prev = NULL, tmp2 = iflist;  tmp2;  prev = tmp2, tmp2 = tmp2->next)
    {
        if (tmp2->channel == channel)
        {
            tmp = tmp2;
            here = 1;
            break;
        }
        /*endif*/
        if (tmp2->channel > channel)
            break;
        /*endif*/
    }
    /*endfor*/

    if (!here)
    {
        if ((tmp = (struct unicall_pvt *) malloc(sizeof(struct unicall_pvt))) == NULL)
        {
            ast_log(LOG_ERROR, "MALLOC FAILED\n");
            return NULL;
        }
        /*endif*/
        memset(tmp, 0, sizeof(struct unicall_pvt));
        for (x = 0;  x < 3;  x++)
            tmp->subs[x].zfd = -1;
        /*endfor*/
        tmp->next = tmp2;
        if (!prev)
            iflist = tmp;
        else
            prev->next = tmp;
        /*endif*/
    }
    /*endif*/

    if (tmp)
    {
        if (channel != CHAN_PSEUDO)
        {
            snprintf(fn, sizeof(fn), "%d", channel);
            /* Open non-blocking */
            if (!here)
                tmp->subs[SUB_REAL].zfd = unicall_open(fn);
            /*endif*/
            if (tmp->subs[SUB_REAL].zfd < 0)
            {
                ast_log(LOG_ERROR, "Unable to open channel %d: %s\nhere = %d, tmp->channel = %d, channel = %d\n", channel, strerror(errno), here, tmp->channel, channel);
                free(tmp);
                return NULL;
            }
            /*endif*/
            memset(&p, 0, sizeof(p));
            if (here  &&  strcmp(tmp->protocol_class, protocol_class) != 0)
            {
                if (reset_channel(tmp))
                {
                    ast_log(LOG_ERROR, "Failed to reset chan_unicall channel %d\n", tmp->channel);
                    return NULL;
                }
                /*endif*/
            }
            /*endif*/
        }
        else
        {
            protocol_class = NULL;
        }
        /*endif*/
        if ((tmp->uc = uc_new(tmp->subs[SUB_REAL].zfd,
                              tmp->subs[SUB_REAL].zfd,
                              protocol_class,
                              protocol_variant,
                              (protocol_end == NULL  ||  strcasecmp(protocol_end, "co") == 0)  ?  UC_MODE_CO  :  UC_MODE_CPE,
                              1)) == NULL)
        {
            ast_log(LOG_WARNING, "Unable to create UC call :(\n");
            unicall_close(tmp->subs[SUB_REAL].zfd);
            free(tmp);
            return NULL;
        }
        /*endif*/
        uc_set_logging(tmp->uc, 0x42, 0, protocol_class);
        uc_set_signaling_callback(tmp->uc, handle_uc_event, (int) tmp);
        uc_set_channel_read_callback(tmp->uc, 0, handle_uc_read, (int) tmp);
        //uc_set_channel_error_callback(tmp->uc, channel_error, (int) tmp);
        uc_Unblock(tmp->uc, -1);
        if (!here  &&  tmp->subs[SUB_REAL].zfd >= 0)
        {
            memset(&bi, 0, sizeof(bi));
            if ((res = ioctl(tmp->subs[SUB_REAL].zfd, ZT_GET_BUFINFO, &bi)))
            {
                ast_log(LOG_WARNING, "Unable to check buffer policy on channel %d\n", channel);
            }
            else
            {
                bi.txbufpolicy = ZT_POLICY_IMMEDIATE;
                bi.rxbufpolicy = ZT_POLICY_IMMEDIATE;
                bi.numbufs = 4;
                bi.bufsize = READ_SIZE;
                if ((res = ioctl(tmp->subs[SUB_REAL].zfd, ZT_SET_BUFINFO, &bi)))
                    ast_log(LOG_WARNING, "Unable to set buffer policy on channel %d\n", channel);
                /*endif*/
            }
            /*endif*/
        }
        /*endif*/
        tmp->immediate = immediate;
        tmp->protocol_class = protocol_class;
        tmp->radio = radio;
        tmp->firstradio = 0;
        /* Flag to destroy the channel must be cleared on new mkif.  Part of changes for reload to work */
        tmp->destroy = 0;
        tmp->threewaycalling = threewaycalling;
        tmp->adsi = adsi;
        tmp->callreturn = callreturn;
        tmp->echocancel = echocancel;
        tmp->echocanbridged = echocanbridged;
        tmp->cancallforward = cancallforward;
        tmp->channel = channel;
        tmp->stripmsd = stripmsd;
        tmp->use_callerid = use_callerid;
        strncpy(tmp->callerid, callerid, sizeof(tmp->callerid) - 1);
        tmp->callerid[sizeof(tmp->callerid) - 1] = '\0';
        strncpy(tmp->accountcode, accountcode, sizeof(tmp->accountcode) - 1);
        tmp->accountcode[sizeof(tmp->accountcode) - 1] = '\0';
        tmp->amaflags = amaflags;
        if (!here)
        {
            tmp->confno = -1;
            tmp->propconfno = -1;
        }
        /*endif*/
        tmp->transfer = transfer;
        ast_mutex_init(&tmp->lock);
        strncpy(tmp->language, language, sizeof(tmp->language) - 1);
        strncpy(tmp->musicclass, musicclass, sizeof(tmp->musicclass) - 1);
        strncpy(tmp->context, context, sizeof(tmp->context) - 1);
        strncpy(tmp->mailbox, mailbox, sizeof(tmp->mailbox) - 1);
        tmp->msgstate = -1;
        tmp->group = cur_group;
        tmp->callgroup=cur_callergroup;
        tmp->pickupgroup=cur_pickupgroup;
        tmp->rxgain = rxgain;
        tmp->txgain = txgain;
        tmp->onhooktime = time(NULL);
        if (tmp->subs[SUB_REAL].zfd >= 0)
        {
            set_actual_gain(tmp->subs[SUB_REAL].zfd, 0, tmp->rxgain, tmp->txgain, tmp->law);
            if (tmp->dsp)
                ast_dsp_digitmode(tmp->dsp, DSP_DIGITMODE_DTMF | tmp->dtmfrelax);
            /*endif*/
            update_conf(tmp);
            tmp->inalarm = 0;
            memset(&si, 0, sizeof(si));
            if (ioctl(tmp->subs[SUB_REAL].zfd, ZT_SPANSTAT, &si))
            {
                ast_log(LOG_ERROR, "Unable to get span status: %s\n", strerror(errno));
                free(tmp);
                return NULL;
            }
            /*endif*/
            if (si.alarms)
                tmp->inalarm = 1;
            /*endif*/
        }
        /*endif*/
    }
    /*endif*/
    return tmp;
}

static inline int available(struct unicall_pvt *p, int channelmatch, int groupmatch)
{
    /* First, check group matching */
    if ((p->group & groupmatch) != groupmatch)
        return 0;
    /*endif*/
    /* Check to see if we have a channel match */
    if (channelmatch > 0  &&  p->channel != channelmatch)
        return 0;
    /*endif*/
    /* If do not disturb, definitely not */
    if (p->dnd)
        return 0;    
    /*endif*/
    /* If no owner definitely available */
    if (p->owner == NULL)
    {
        if (p->uc)
        {
            if (p->crn  ||  p->blocked)
                return 0;
            /*endif*/
            return 1;
        }
        /*endif*/
        if (!p->radio)
        {
            /* Check hook state */
            /* TODO: check hook state */
        }
        /*endif*/
        return 1;
    }
    /*endif*/

    return 0;
}

static struct unicall_pvt *chandup(struct unicall_pvt *src)
{
    struct unicall_pvt *p;
    ZT_BUFFERINFO bi;

    if ((p = malloc(sizeof(struct unicall_pvt))) == NULL)
        return NULL;
    /*endif*/
    memcpy(p, src, sizeof(struct unicall_pvt));
    /* Allocate a zapata structure */
    if ((p->subs[SUB_REAL].zfd = unicall_open("/dev/zap/pseudo")) < 0)
    {
        ast_log(LOG_ERROR, "Unable to dup channel: %s\n",  strerror(errno));
        free(p);
        return NULL;
    }
    /*endif*/
    if (ioctl(p->subs[SUB_REAL].zfd, ZT_GET_BUFINFO, &bi))
    {
        ast_log(LOG_WARNING, "Unable to check buffer policy on dup channel\n");
    }
    else
    {
        bi.txbufpolicy = ZT_POLICY_IMMEDIATE; 
        bi.rxbufpolicy = ZT_POLICY_IMMEDIATE;
        bi.numbufs = 4;
        bi.bufsize = READ_SIZE;
        if (ioctl(p->subs[SUB_REAL].zfd, ZT_SET_BUFINFO, &bi))
            ast_log(LOG_WARNING, "Unable to set buffer policy on dup channel\n");
        /*endif*/
    }
    /*endif*/
    p->destroy = 1;
    p->next = iflist;
    iflist = p;
    return p;
}
    
static struct ast_channel *unicall_request(char *type, int format, void *data)
{
    struct unicall_pvt *p;
    int oldformat;
    int groupmatch = 0;
    int channelmatch = -1;
    int callwait = 0;
    struct ast_channel *tmp = NULL;
    char *dest = NULL;
    int x;
    char *s;
    char opt = 0;
    int res = 0;
    int y = 0;
    
    /* We do signed linear */
    oldformat = format;
    format &= (AST_FORMAT_SLINEAR | AST_FORMAT_ALAW | AST_FORMAT_ULAW);
    if (!format)
    {
        ast_log(LOG_NOTICE, "Asked to get a channel of unsupported format '%d'\n", oldformat);
        return NULL;
    }
    /*endif*/
    if (data == NULL)
    {
        ast_log(LOG_WARNING, "Channel requested with no data\n");
        return NULL;
    }
    /*endif*/
    if ((dest = strdup((char *) data)) == NULL)
    {
        ast_log(LOG_ERROR, "Out of memory\n");
        return NULL;
    }
    /*endif*/
    if (dest[0] == 'g')
    {
        /* Retrieve the group number */
        char *stringp = NULL;

        stringp = dest + 1;
        s = strsep(&stringp, "/");
        if ((res = sscanf(s, "%d%c%d", &x, &opt, &y)) < 1)
        {
            ast_log(LOG_WARNING, "Unable to determine group for data %s\n", (char *) data);
            free(dest);
            return NULL;
        }
        /*endif*/
        groupmatch = 1 << x;
    }
    else
    {
        char *stringp = NULL;

        stringp = dest;
        s = strsep(&stringp, "/");
        if (!strcasecmp(s, "pseudo"))
        {
            /* Special case for pseudo */
            x = CHAN_PSEUDO;
        }
        else if ((res = sscanf(s, "%d%c%d", &x, &opt, &y)) < 1)
        {
            ast_log(LOG_WARNING, "Unable to determine channel for data %s\n", (char *) data);
            free(dest);
            return NULL;
        }
        /*endif*/
        channelmatch = x;
    }
    /*endif*/
    /* Search for an unowned channel */
    if (ast_mutex_lock(&iflock))
    {
        ast_log(LOG_ERROR, "Unable to lock interface list???\n");
        return NULL;
    }
    /*endif*/
    for (p = iflist;  p  &&  tmp == NULL;  p = p->next)
    {
        if (available(p, channelmatch, groupmatch))
        {
            if (option_debug)
                ast_log(LOG_DEBUG, "Using channel %d\n", p->channel);
            /*endif*/
            if (p->reserved  ||  p->inalarm)
                continue;
            /*endif*/
            if (p->uc)
            {
#if 0
                if ((p->call = uc_new_call(p->uc, NULL)) == NULL)
                {
                    ast_log(LOG_WARNING, "Unable to create call on channel %d\n", p->channel);
                    break;
                }
                /*endif*/
#endif
            }
            /*endif*/
            callwait = (p->owner != NULL);
            if (p->channel == CHAN_PSEUDO)
            {
                if ((p = chandup(p)) == NULL)
                    break;
                /*endif*/
            }
            /*endif*/
            if (p->owner)
            {
                if (alloc_sub(p, SUB_CALLWAIT))
                {
                    p = NULL;
                    break;
                }
                /*endif*/
            }
            /*endif*/
            tmp = unicall_new(p, AST_STATE_RESERVED, 0, p->owner  ?  SUB_CALLWAIT  :  SUB_REAL, 0);
            /* Make special notes */
            if (res > 1)
            {
                if (opt == 'c')
                {
                    /* Confirm answer */
                    p->confirmanswer = 1;
                }
                else if (opt == 'r')
                {
                    /* Distinctive ring */
                    if (res < 3)
                        ast_log(LOG_WARNING, "Distinctive ring missing identifier in '%s'\n", (char *) data);
                    else
                        p->distinctivering = y;
                    /*endif*/
                }
                else if (opt == 'd')
                {
                    /* If this is an ISDN call, make it digital */
                    p->digital = 1;
                }
                else
                {
                    ast_log(LOG_WARNING, "Unknown option '%c' in '%s'\n", opt, (char *) data);
                }
                /*endif*/
            }
            /*endif*/
            /* Note if the call is a call waiting call */
            if (callwait)
                tmp->cdrflags |= AST_CDR_CALLWAIT;
            /*endif*/
            break;
        }
        /*endif*/
    }
    /*endfor*/
    ast_mutex_unlock(&iflock);
    restart_monitor();
    return tmp;
}

static int get_group(char *s)
{
    char *copy;
    char *piece;
    char *c = NULL;
    int start = 0;
    int finish = 0;
    int x;
    int group = 0;

    if ((copy = strdup(s)) == NULL)
    {
        ast_log(LOG_ERROR, "Out of memory\n");
        return 0;
    }
    /*endif*/
    c = copy;
    piece = strsep(&c, ",");
    while (piece)
    {
        if (sscanf(piece, "%d-%d", &start, &finish) == 2)
        {
            /* Range */
        }
        else if (sscanf(piece, "%d", &start))
        {
            /* Just one */
            finish = start;
        }
        else
        {
            ast_log(LOG_ERROR, "Syntax error parsing '%s' at '%s'.  Using '0'\n", s,piece);
            return 0;
        }
        piece = strsep(&c, ",");
        for (x = start;  x <= finish;  x++)
        {
            if (x > 31  ||  x < 0)
                ast_log(LOG_WARNING, "Ignoring invalid group %d\n", x);
            else
                group |= (1 << x);
            /*endif*/
        }
        /*endfor*/
    }
    /*endwhile*/
    free(copy);
    return group;
}

static char *complete_span(char *line, char *word, int pos, int state)
{
    int span;
    char tmp[50];

    for (span = 1;  span <= NUM_SPANS;  span++)
    {
        if (span > state)
        {
            snprintf(tmp, sizeof(tmp), "%d", span);
            return strdup(tmp);
        }
        /*endif*/
        span++;
    }
    /*endfor*/
    return NULL;
}

static int handle_uc_debug(int fd, int argc, char *argv[])
{
    int chan;
    struct unicall_pvt *tmp = NULL;

    if (argc < 4)
        return RESULT_SHOWUSAGE;
    /*endif*/
    chan = atoi(argv[3]);
    if (chan < 1  ||  chan > NUM_SPANS)
    {
        ast_cli(fd, "Invalid span %s.  Should be a number %d to %d\n", argv[3], 1, NUM_SPANS);
        return RESULT_SUCCESS;
    }
    /*endif*/
    for (tmp = iflist;  tmp;  tmp = tmp->next)
    {
        if (tmp->channel == chan)
        {
            if (tmp->uc)
            {
                uc_set_logging(tmp->uc, 0x7FFFFFFF, 0, "XXX");
                ast_cli(fd, "Enabled debugging on channel %d\n", chan);
                return RESULT_SUCCESS;
            }
            /*endif*/
            break;
        }
        /*endif*/
    }
    /*endfor*/
    if (tmp) 
        ast_cli(fd, "UniCall not running on channel %d\n", chan);
    else
        ast_cli(fd, "No such zap channel %d\n", chan);
    /*endif*/
    return RESULT_SUCCESS;
}

static int handle_uc_no_debug(int fd, int argc, char *argv[])
{
    int chan;
    struct unicall_pvt *tmp = NULL;

    if (argc < 5)
        return RESULT_SHOWUSAGE;
    /*endif*/
    chan = atoi(argv[4]);
    if (chan < 1  ||  chan > NUM_SPANS)
    {
        ast_cli(fd, "Invalid channel %d.  Should be a number greater than 0\n", chan);
        return RESULT_SUCCESS;
    }
    /*endif*/
    for (tmp = iflist;  tmp;  tmp = tmp->next)
    {
        if (tmp->channel == chan)
        {
            if (tmp->uc)
            {
                uc_set_logging(tmp->uc, 0, 0, "");
                ast_cli(fd, "Disabled debugging on channel %d\n", chan);
                return RESULT_SUCCESS;
            }
            /*endif*/
            break;
        }
        /*endif*/
    }
    /*endwhile*/
    if (tmp) 
        ast_cli(fd, "UniCall not running on channel %d\n", chan);
    else
        ast_cli(fd, "No such zap channel %d\n", chan);
    /*endif*/
    return RESULT_SUCCESS;
}

static char uc_debug_help[] = 
    "Usage: UC debug span <span>\n"
    "       Enables debugging on a given PRI span\n";
    
static char uc_no_debug_help[] = 
    "Usage: UC no debug span <span>\n"
    "       Disables debugging on a given PRI span\n";

static struct ast_cli_entry uc_debug =
{
    { "UC", "debug", "span", NULL }, handle_uc_debug, "Enables UC debugging on a span", uc_debug_help, complete_span 
};

static struct ast_cli_entry uc_no_debug =
{
    { "UC", "no", "debug", "span", NULL }, handle_uc_no_debug, "Disables PRI debugging on a span", uc_no_debug_help, complete_span
};

static int unicall_destroy_channel(int fd, int argc, char **argv)
{
    struct unicall_pvt *tmp;
    struct unicall_pvt *prev;
    int channel;
    
    if (argc != 4)
        return RESULT_SHOWUSAGE;
    /*endif*/
    channel = atoi(argv[3]);

    for (prev = NULL, tmp = iflist;  tmp;  prev = tmp, tmp = tmp->next)
    {
        if (tmp->channel == channel)
        {
            destroy_channel(prev, tmp, 1);
            return RESULT_SUCCESS;
        }
        /*endif*/
    }
    /*endfor*/
    return RESULT_FAILURE;
}

static int unicall_show_channels(int fd, int argc, char **argv)
{
#define FORMAT "%3d %-10.10s %-10.10s %-10.10s %-8.8s\n"
#define FORMAT2 "%3s %-10.10s %-10.10s %-10.10s %-8.8s\n"
    struct unicall_pvt *tmp;

    if (argc != 3)
        return RESULT_SHOWUSAGE;
    /*endif*/
    ast_mutex_lock(&iflock);
    ast_cli(fd, FORMAT2, "Chan. Num.", "Extension", "Context", "Language", "MusicOnHold");
    
    for (tmp = iflist;  tmp;  tmp = tmp->next)
        ast_cli(fd, FORMAT, tmp->channel, tmp->exten, tmp->context, tmp->language, tmp->musicclass);
    /*endfor*/
    ast_mutex_unlock(&iflock);
    return RESULT_SUCCESS;
#undef FORMAT
#undef FORMAT2
}

static int unicall_show_channel(int fd, int argc, char **argv)
{
    struct unicall_pvt *tmp = NULL;
    int channel;
    int x;

    if (argc != 4)
        return RESULT_SHOWUSAGE;
    /*endif*/
    channel = atoi(argv[3]);

    ast_mutex_lock(&iflock);
    for (tmp = iflist;  tmp;  tmp = tmp->next)
    {
        if (tmp->channel == channel)
        {
            ast_cli(fd, "Channel: %d\n", tmp->channel);
            ast_cli(fd, "File descriptor: %d\n", tmp->subs[SUB_REAL].zfd);
            ast_cli(fd, "Span: %d\n", tmp->span);
            ast_cli(fd, "Extension: %s\n", tmp->exten);
            ast_cli(fd, "Context: %s\n", tmp->context);
            ast_cli(fd, "Destroy: %d\n", tmp->destroy);
            ast_cli(fd, "Signalling type: %s\n", tmp->protocol_class);
            ast_cli(fd, "Signalling variant: %s\n", tmp->protocol_variant);
            ast_cli(fd, "Owner: %s\n", tmp->owner  ?  tmp->owner->name  :  "<None>");
            ast_cli(fd, "Real: %s%s\n", tmp->subs[SUB_REAL].owner ? tmp->subs[SUB_REAL].owner->name : "<None>", tmp->subs[SUB_REAL].inthreeway ? " (Confed)" : "");
            ast_cli(fd, "Callwait: %s%s\n", tmp->subs[SUB_CALLWAIT].owner ? tmp->subs[SUB_CALLWAIT].owner->name : "<None>", tmp->subs[SUB_CALLWAIT].inthreeway ? " (Confed)" : "");
            ast_cli(fd, "Threeway: %s%s\n", tmp->subs[SUB_THREEWAY].owner ? tmp->subs[SUB_THREEWAY].owner->name : "<None>", tmp->subs[SUB_THREEWAY].inthreeway ? " (Confed)" : "");
            ast_cli(fd, "Confno: %d\n", tmp->confno);
            ast_cli(fd, "Propagated conference: %d\n", tmp->propconfno);
            ast_cli(fd, "Real in conference: %d\n", tmp->inconference);
            ast_cli(fd, "Dialing: %d\n", tmp->dialing);
            ast_cli(fd, "Default law: %s\n", (tmp->law == UC_CODEC_ULAW)  ?  "ulaw"  :  tmp->law == UC_CODEC_ALAW  ?  "alaw"  :  "unknown");
            ast_cli(fd, "Fax handled: %s\n", tmp->faxhandled ? "yes" : "no");
            if (tmp->master)
                ast_cli(fd, "Master Channel: %d\n", tmp->master->channel);
            /*endif*/
            for (x = 0;  x < MAX_SLAVES;  x++)
            {
                if (tmp->slaves[x])
                    ast_cli(fd, "Slave Channel: %d\n", tmp->slaves[x]->channel);
                /*endif*/
            }
            /*endfor*/
            if (tmp->uc)
            {
                ast_cli(fd, "UniCall flags: ");
                if (tmp->blocked)
                    ast_cli(fd, "Blocked ");
                /*endif*/
                if (tmp->crn)
                    ast_cli(fd, "Call ");
                /*endif*/
                ast_cli(fd, "\n");
            }
            /*endif*/
            ast_mutex_unlock(&iflock);
            return RESULT_SUCCESS;
        }
        /*endif*/
    }
    /*endfor*/
    
    ast_cli(fd, "Unable to find given channel %d\n", channel);
    ast_mutex_unlock(&iflock);
    return RESULT_FAILURE;
}

static char show_channels_usage[] =
    "Usage: UC show channels\n"
    "    Shows a list of available channels\n";

static char show_channel_usage[] =
    "Usage: UC show channel <chan num>\n"
    "    Detailed information about a given channel\n";
static char destroy_channel_usage[] =
    "Usage: UC destroy channel <chan num>\n"
    "    DON'T USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING.  Immediately removes a given channel, whether it is in use or not\n";

static struct ast_cli_entry cli_show_channels =
{ 
    {"UC", "show", "channels", NULL}, unicall_show_channels, "Show active UniCall channels", show_channels_usage, NULL
};

static struct ast_cli_entry cli_show_channel =
{ 
    {"UC", "show", "channel", NULL}, unicall_show_channel, "Show information on a channel", show_channel_usage, NULL
};

static struct ast_cli_entry cli_destroy_channel =
{ 
    {"UC", "destroy", "channel", NULL}, unicall_destroy_channel, "Destroy a channel", destroy_channel_usage, NULL
};

int load_module(void)
{
    struct unicall_pvt *tmp;
    struct ast_config *cfg;
    struct ast_variable *v;
    char *chan;
    char *c;
    int start;
    int finish;
    int x;
    int y;
    int cur_radio = 0;

    uc_Start();
    uc_set_error_handler(unicall_error);
    uc_set_message_handler(unicall_message);

    /* We *must* have a config file, otherwise stop immediately */
    if ((cfg = ast_load(config)) == NULL)
    {
        ast_log(LOG_ERROR, "Unable to load config %s\n", config);
        return -1;
    }
    /*endif*/

    if (ast_mutex_lock(&iflock))
    {
        /* It's a little silly to lock it, but we might as well just to be sure */
        ast_log(LOG_ERROR, "Unable to lock interface list???\n");
        return -1;
    }
    /*endif*/
    cur_protocol_class = NULL;
    cur_protocol_variant = NULL;
    cur_protocol_end = NULL;

    /* Set some defaults for things not specified in the config file. */
    use_callerid = 1;

    v = ast_variable_browse(cfg, "channels");
    while (v)
    {
        /* Create the interface list */
        if (strcasecmp(v->name, "channel") == 0)
        {
            if (cur_protocol_class == NULL)
            {
                ast_log(LOG_ERROR, "Signalling must be specified before any channels are.\n");
                ast_destroy(cfg);
                ast_mutex_unlock(&iflock);
                unload_module();
                return -1;
            }
            /*endif*/
            c = v->value;
            chan = strsep(&c, ",");
            while (chan)
            {
                if (sscanf(chan, "%d-%d", &start, &finish) == 2)
                {
                    /* Range */
                }
                else if (sscanf(chan, "%d", &start))
                {
                    /* Just one */
                    finish = start;
                }
                else if (!strcasecmp(chan, "pseudo"))
                {
                    finish = start = CHAN_PSEUDO;
                }
                else
                {
                    ast_log(LOG_ERROR, "Syntax error parsing '%s' at '%s'\n", v->value, chan);
                    ast_destroy(cfg);
                    ast_mutex_unlock(&iflock);
                    unload_module();
                    return -1;
                }
                /*endif*/
                if (finish < start)
                {
                    ast_log(LOG_WARNING, "Sillyness: %d < %d\n", start, finish);
                    x = finish;
                    finish = start;
                    start = x;
                }
                /*endif*/
                for (x = start;  x <= finish;  x++)
                {
                    if ((tmp = mkintf(x, cur_protocol_class, cur_protocol_variant, cur_protocol_end, cur_radio)) == 0)
                    {
                        ast_log(LOG_ERROR, "Unable to register channel '%s'\n", v->value);
                        ast_destroy(cfg);
                        ast_mutex_unlock(&iflock);
                        unload_module();
                        return -1;
                    }
                    /*endif*/
                    if (option_verbose > 2)
                        ast_verbose(VERBOSE_PREFIX_3 "Registered channel %d, %s signalling\n", x, tmp->protocol_class);
                    /*endif*/
                }
                /*endfor*/
                chan = strsep(&c, ",");
            }
            /*endwhile*/
        }
        else if (!strcasecmp(v->name, "protocolclass"))
        {
            cur_protocol_class = strdup(v->value);
	    cur_radio = 0;
        }
        else if (!strcasecmp(v->name, "protocolvariant"))
        {
            cur_protocol_variant = strdup(v->value);
	    cur_radio = 0;
        }
        else if (!strcasecmp(v->name, "protocolend"))
        {
            cur_protocol_end = strdup(v->value);
	    cur_radio = 0;
        }
        else if (strcasecmp(v->name, "threewaycalling") == 0)
        {
            threewaycalling = ast_true(v->value);
        }
        else if (strcasecmp(v->name, "cancallforward") == 0)
        {
            cancallforward = ast_true(v->value);
        }
        else if (strcasecmp(v->name, "mailbox") == 0)
        {
            strncpy(mailbox, v->value, sizeof(mailbox) -1);
        }
        else if (strcasecmp(v->name, "adsi") == 0)
        {
            adsi = ast_true(v->value);
        }
        else if (strcasecmp(v->name, "transfer") == 0)
        {
            transfer = ast_true(v->value);
        }
        else if (strcasecmp(v->name, "echocancelwhenbridged") == 0)
        {
            echocanbridged = ast_true(v->value);
        }
        else if (strcasecmp(v->name, "echocancel") == 0)
        {
            if (v->value  &&  strlen(v->value))
                y = atoi(v->value);
            else
                y = 0;
            /*endif*/
            if ((y == 32)  ||  (y == 64)  ||  (y == 128)  ||  (y == 256))
                echocancel = y;
            else
                echocancel = ast_true(v->value);
            /*endif*/
        }
        else if (strcasecmp(v->name, "callreturn") == 0)
        {
            callreturn = ast_true(v->value);
        }
        else if (strcasecmp(v->name, "context") == 0)
        {
            strncpy(context, v->value, sizeof(context) - 1);
        }
        else if (strcasecmp(v->name, "language") == 0)
        {
            strncpy(language, v->value, sizeof(language) - 1);
        }
        else if (strcasecmp(v->name, "musiconhold") == 0)
        {
            strncpy(musicclass, v->value, sizeof(musicclass) - 1);
        } 
        else if (strcasecmp(v->name, "stripmsd") == 0)
        {
            stripmsd = atoi(v->value);
        } 
        else if (strcasecmp(v->name, "group") == 0)
        {
            cur_group = get_group(v->value);
        } 
        else if (strcasecmp(v->name, "callgroup") == 0)
        {
            cur_callergroup = get_group(v->value);
        } 
        else if (strcasecmp(v->name, "pickupgroup") == 0)
        {
            cur_pickupgroup = get_group(v->value);
        } 
        else if (strcasecmp(v->name, "immediate") == 0)
        {
            immediate = ast_true(v->value);
        } 
        else if (strcasecmp(v->name, "rxgain") == 0)
        {
            if (sscanf(v->value, "%f", &rxgain) != 1)
                ast_log(LOG_WARNING, "Invalid rxgain: %s\n", v->value);
            /*endif*/
        } 
        else if (strcasecmp(v->name, "txgain") == 0)
        {
            if (sscanf(v->value, "%f", &txgain) != 1)
                ast_log(LOG_WARNING, "Invalid txgain: %s\n", v->value);
            /*endif*/
        } 
        else if (strcasecmp(v->name, "accountcode") == 0)
        {
            strncpy(accountcode, v->value, sizeof(accountcode) - 1);
        } 
        else if (strcasecmp(v->name, "amaflags") == 0)
        {
            y = ast_cdr_amaflags2int(v->value);
            if (y < 0) 
                ast_log(LOG_WARNING, "Invalid AMA flags: %s at line %d\n", v->value, v->lineno);
            else
                amaflags = y;
            /*endif*/
        } 
        else if (strcasecmp(v->name, "minunused") == 0)
        {
            minunused = atoi(v->value);
        }
        else if (strcasecmp(v->name, "idleext") == 0)
        {
            strncpy(idleext, v->value, sizeof(idleext) - 1);
        }
        else if (strcasecmp(v->name, "idledial") == 0)
        {
            strncpy(idledial, v->value, sizeof(idledial) - 1);
        }
        else if (strcasecmp(v->name, "usecallerid") == 0)
        {
            use_callerid = ast_true(v->value);
        }
	else if (!strcasecmp(v->name, "callerid"))
        {
            if (!strcasecmp(v->value, "asreceived"))
                callerid[0] = '\0';
            else
                strncpy(callerid, v->value, sizeof(callerid) - 1);
            /*endif*/
            callerid[sizeof(callerid) - 1] = '\0';
        }
        else
        {
            ast_log(LOG_DEBUG, "Ignoring %s\n", v->name);
        }
        /*endif*/
        v = v->next;
    }
    /*endwhile*/
    ast_mutex_unlock(&iflock);
    /* Make sure we can register our UC channel type */
    if (ast_channel_register(type, tdesc, AST_FORMAT_SLINEAR | AST_FORMAT_ALAW | AST_FORMAT_ULAW, unicall_request))
    {
        ast_log(LOG_ERROR, "Unable to register channel class %s\n", type);
        ast_destroy(cfg);
        unload_module();
        return -1;
    }
    /*endif*/
    if (ast_channel_register(typecompat, tdesc, AST_FORMAT_SLINEAR | AST_FORMAT_ALAW | AST_FORMAT_ULAW, unicall_request))
    {
        ast_log(LOG_ERROR, "Unable to register channel class %s\n", typecompat);
        ast_destroy(cfg);
        unload_module();
        return -1;
    }
    /*endif*/
    ast_destroy(cfg);

    ast_cli_register(&uc_debug);
    ast_cli_register(&uc_no_debug);

    ast_cli_register(&cli_show_channels);
    ast_cli_register(&cli_show_channel);
    ast_cli_register(&cli_destroy_channel);
    /* And start the monitor for the first time */
    restart_monitor();
    return 0;
}

int unload_module(void)
{
    struct unicall_pvt *p;
    struct unicall_pvt *pl;

    /* First, take us out of the channel loop */
    ast_channel_unregister(type);
    ast_channel_unregister(typecompat);
    ast_cli_unregister(&cli_show_channels);
    ast_cli_unregister(&cli_show_channel);
    ast_cli_unregister(&cli_destroy_channel);
    if (ast_mutex_lock(&iflock))
    {
        ast_log(LOG_WARNING, "Unable to lock the monitor\n");
        return -1;
    }
    /*endif*/
    /* Hangup all interfaces if they have an owner */
    for (p = iflist;  p;  p = p->next)
    {
        if (p->owner)
            ast_softhangup(p->owner, AST_SOFTHANGUP_APPUNLOAD);
        /*endif*/
    }
    /*endfor*/
    iflist = NULL;
    ast_mutex_unlock(&iflock);
    if (ast_mutex_lock(&monlock))
    {
        ast_log(LOG_WARNING, "Unable to lock the monitor\n");
        return -1;
    }
    /*endif*/
    if (monitor_thread)
    {
        pthread_cancel(monitor_thread);
        pthread_kill(monitor_thread, SIGURG);
        pthread_join(monitor_thread, NULL);
    }
    /*endif*/
    monitor_thread = -2;
    ast_mutex_unlock(&monlock);

    if (ast_mutex_lock(&iflock))
    {
        ast_log(LOG_WARNING, "Unable to lock the monitor\n");
        return -1;
    }
    /*endif*/
    /* Destroy all the interfaces and free their memory */
    for (p = iflist;  p;  )
    {
        /* Close the zapata thingy */
        if (p->subs[SUB_REAL].zfd >= 0)
            unicall_close(p->subs[SUB_REAL].zfd);
        /*endif*/
        pl = p;
        p = p->next;
        /* Free associated memory */
        free(pl);
    }
    /*endfor*/
    iflist = NULL;
    ast_mutex_unlock(&iflock);
    return 0;
}

static int unicall_sendtext(struct ast_channel *c, char *text)
{
    struct unicall_pvt *p = c->pvt->pvt;
    int index;

    if ((index = unicall_get_index(c, p, 0)) < 0)
    {
        ast_log(LOG_WARNING, "Huh?  I don't exist?\n");
        return -1;
    }
    /*endif*/
    if (text[0] == '\0')
        return 0;
    /*endif*/
    /* TODO: */
    //uc_SendUserUserText(p->uc, text);
    return  -1;
}

int usecount(void)
{
    int res;

    ast_mutex_lock(&usecnt_lock);
    res = usecnt;
    ast_mutex_unlock(&usecnt_lock);
    return res;
}

char *description(void)
{
    return desc;
}

char *key(void)
{
    return ASTERISK_GPL_KEY;
}
