<?php
/**
 * this file contains the root class 
 * SocketLayer.
 * @package netsock
 */

/**
 * require logger class
 */
require_once 'stdlog/Logger.php';
require_once 'SocketException.php';
require_once 'BindException.php';
require_once 'ConnectException.php';
require_once 'InvalidSocketResourceException.php';

/**
 * SocketLayer
 *
 * This is the root class of the API, common default needed values
 */
abstract class SocketLayer
{
	/**
	 * string socket types, used in socket url resources
	 */
	const SOCKET_TYPE_TCP           = 'tcp';
	const SOCKET_TYPE_UDP           = 'udp';
	const SOCKET_TYPE_UNIX          = 'unix';

	/**
	 * how many bytes to read from the socket
	 */
	const SOCKET_READ_BYTES         = 2048;
	
	/**
	 * predefined socket status
	 */
	const SOCKET_STATUS_CONNECTING  = 115;
	
	/**
	 * predefined socket status
	 */
	const SOCKET_STATUS_IN_PROGRESS = 114;

	/**
	 * recognized socket resource string
	 * regular expression
	 */
	const SOCKET_RESOURCE_REGEXP    = '^(tcp|udp|unix)://([/.A-Za-z0-9_-]*)(:[0-9]+)?$';
	
	/**
	 * php date format to use in all the log and debug operations
	 * @var string
	 */
	protected static $date_format   = 'l dS \of F Y h:i:s A';

	/**
	 * main socket where we write/read to/from
	 * @var resource
	 */
	protected $main_socket          = NULL;
	
	/**
	 * socket URI string that tell us protocol, server and other usefull stuff
	 * @var string
	 */
	protected $socket_uri           = '';

	/**
	 * @return void
	 * @param String $SocketResource
	 * @throw InvalidSocketResourceException
	 */
	public static function ParseSocketResource($SocketResource)
	{
		$arguments = array();
		/* Extracting socket type, address and optional port from resource string */
		eregi(SocketLayer::SOCKET_RESOURCE_REGEXP, $SocketResource, $arguments);
		/* making sure the resource is at least valid */
		if ( 4 != count($arguments) )
		{
			$log_message = "hu?, i guess this is not a valid socket resource {$SocketResource}";
			Logger::WriteLog($log_message, Logger::LOG_LEVEL_FATAL);
			throw new InvalidSocketResourceException($log_message);
		}
		if ( FALSE !== $arguments[3] )
		{
			$arguments[3] = substr($arguments[3], 1);
		}
		/* return everything except the whole string */
		return array_slice($arguments, 1);
	}
	
	/**
	 * @return void
	 * @param string $SocketResource
	 * @throw InvalidSocketResourceException, SocketException
	 */
	protected static function SocketCreate($SocketResource)
	{
		list($socket_type, $socket_address, $socket_port) = self::ParseSocketResource($SocketResource);
		Logger::WriteLog("Attempting to create socket {$socket_type}, at {$socket_address}:{$socket_port}", Logger::LOG_LEVEL_DEBUG);
		$family   = ( self::SOCKET_TYPE_TCP == $socket_type || self::SOCKET_TYPE_UDP  == $socket_type ) ? AF_INET     : AF_UNIX;
		$type     = ( self::SOCKET_TYPE_TCP == $socket_type || self::SOCKET_TYPE_UNIX == $socket_type ) ? SOCK_STREAM : SOCK_DGRAM;
		$protocol = ( AF_INET == $family ) ? SOL_TCP : 0;
		/* Creating socket */
		$socket        = socket_create($family, $type, $protocol);
		$socket_status = socket_last_error($socket);
		if ( FALSE === $socket )
		{
			 $socket_error = "socket_create() ({$socket_status}) failed: " . socket_strerror($socket);
			 Logger::WriteLog($socket_error, Logger::LOG_LEVEL_FATAL);
			 throw new SocketException($socket_error);
		}
		return $socket;
	}

}
?>
