<?php
/**
 * this file contains the SocketServer class
 * @package netsock
 */

/**
 * requires the file Socketlayer.php
 * that file contains the abstract root
 * class of the API
 */
require_once 'SocketLayer.php';

/**
 * SocketServer
 * this class is an very high abstraction of a TCP/Unix Socket Server.
 * @abstract
 */
abstract class SocketServer extends SocketLayer
{
	
	/**
	 * Saves the clients accepted sockets resources
	 * in the form $this->clients_sockets[$index] = $client_socket_resource;
	 * @var array
	 */
	private $clients_sockets   = array();
	
	/**
	 * Saves the clients indexes related to client socket resource as key
	 * in the form $this->clients_indexes[$client_socket_resource] = $client_index;
	 * @var array
	 */
	private $clients_indexes   = array();
	
	/**
	 * it keeps track of the indexes that are available to assign to incoming client connections
	 * @var array
	 */
	private $available_indexes = array();
	
	/**
	 * array of socket resources where we should wait for data to retrieve
	 * @var array
	 */
	private $watched_sockets   = array();
	
	/**
	 * array of socket resources that have data ready to be read
	 * @var array
	 */
	private $readable_sockets  = array();
	
	/**
	 * number of maximum allowed clients
	 * @var int
	 */
	private $max_clients       = 500;
	
	/**
	 * wheter or not we are listening for incoming connections
	 * @var bool
	 */
	private $listening         = FALSE;
	
	/**
	 * socket_select timeout in seconds
	 * @var int
	 */
	private $socket_tvsec      = NULL;
	
	/**
	 * tells what to do when the server starts
	 * @param void
	 * @return void
	 */
	protected function OnStart(){}

	/**
	 * tells what to do when data from clients is available, but before the data is read
	 * @param void
	 * @return void
	 */
	protected function OnDataAvailable(){}
	
	/**
	 * tells what to do when some client connects to the server
	 * @param int $ClientIndex
	 * @return void
	 */
	protected function OnConnect($ClientIndex){}
	
	/**
	 * tells what to do when some client disconnects from the server
	 * @param int $ClientIndex
	 * @return void
	 */
	protected function OnDisconnect($ClientIndex){}
	
	/**
	 * tells what to do when data is received from some client
	 * @param int $ClientIndex
	 * @param string $ClientData
	 * @return void
	 */
	protected function OnClientDataReceipt($ClientIndex, $ClientData){}
	
	/**
	 * tells what to do when data is received from some client
	 * @param void
	 * @return void
	 */
	protected static function LogServerMemoryUsage()
	{
		if ( function_exists('memory_get_usage') )
		{
			$memory_usage = memory_get_usage();
			$memory_usage = (($memory_usage / 1024) / 1024); /* MB */
			$memory_usage = number_format($memory_usage, 2) . ' MB';
			$date         = date(SocketServer::$date_format);
			$memory_usage = "memory usage at {$date} is: {$memory_usage}";
			Logger::WriteLog($memory_usage, Logger::LOG_LEVEL_DEBUG);
		}
		else
		{
			Logger::WriteLog("unable to get memory usage, compile php with --enable-memory-limit");
		}
	}
	
	/**
	 * set the amount of time to wait in socket_select, NULL for no timeout
	 * @param int $TimeoutSeconds
	 * @return void
	 */
	protected function SetSocketTvsec($TimeoutSeconds)
	{
		$this->socket_tvsec = $TimeoutSeconds;
	}
	
	/**
	 * start to listen for incoming connections
	 * @param string $SocketUri
	 * @return void
	 * @throw SocketException, BindException, InvalidSocketResourceException, ListenException
	 */
	protected function ListenConnections($SocketUri)
	{
		$this->socket_uri        = $SocketUri;
		$this->main_socket       = SocketLayer::SocketCreate($SocketUri);
		socket_set_option($this->main_socket, SOL_SOCKET, SO_REUSEADDR, 1);
		self::SocketBind($this->main_socket, $SocketUri);
		if ( !socket_listen($this->main_socket, $this->max_clients) )
		{
			$socket_error = socket_strerror(socket_last_error($this->main_socket));
			Logger::WriteLog("Could not listen ({$socket_error})");
			throw new ListenException($socket_error);
		}
		$date = date(SocketServer::$date_format);
		Logger::WriteLog("Listening... Server started on {$date}.");
		$this->listening = TRUE;
		/* We must add initial socket to watched sockets array, so we know when we have an incoming connection
		 (when initial socket claims to have data on it) */
		$this->WatchSocket($this->main_socket);
	}

	/**
	 * bind a socket acording to the socket uri properties
	 * @return void
	 * @param resource $Socket
	 * @param string $SocketUri
	 * @throw BindException, InvalidSocketResourceException
	 */
	protected static function SocketBind($Socket, $SocketUri)
	{
		list($socket_type, $socket_address, $socket_port) = SocketLayer::ParseSocketResource($SocketUri);
		if ( SocketLayer::SOCKET_TYPE_UNIX == $socket_type )
		{
			unlink($socket_address);
			if ( socket_bind($Socket, $socket_address) )
			{
				return;
			}
		}
		else
		{
			if ( socket_bind($Socket, $socket_address, $socket_port) )
			{
				return;	
			}
		}
		$socket_error = 'Could not bind socket: ' . socket_strerror(socket_last_error($Socket));
		Logger::WriteLog($socket_error, Logger::LOG_LEVEL_FATAL);
		throw new BindException($socket_error);
	}
	
	/**
	 * wait for incoming data until some data arrives
	 * @param void
	 * @return int
	 */
	protected function WaitForData()
	{
		SocketServer::LogServerMemoryUsage();
		$this->readable_sockets = $this->watched_sockets;
		$tvsec_string = $this->socket_tvsec === NULL ? 'NULL' : $this->socket_tvsec;
		Logger::WriteLog('calling socket_select() with tv_sec of ' . $tvsec_string, Logger::LOG_LEVEL_DEBUG);
		$ready                  = socket_select($this->readable_sockets, $write = NULL, $except = NULL, $this->socket_tvsec);
		Logger::WriteLog('returning from socket_select()', Logger::LOG_LEVEL_DEBUG);
		if ( $ready === FALSE )
		{
			Logger::WriteLog('socket_select failed');
		}
		/* If initial socket has data, then it means that we have a new connection, 
		   so we accept it (No limits impposed) */
		if ( $this->SocketHasData($this->main_socket) )
		{
			$this->AcceptConnection();
		}
		return $ready;
	}
	
	/**
	 * how many clients are connected
	 * @param void
	 * @return int
	 * @access public
	 */
	public function GetClientsNumber()
	{
		return count($this->clients_sockets);
	}
	
	/**
	 * whether or not a client exists
	 * @param int $ClientNumber
	 * @return bool
	 * @access public
	 */
	public function ClientExists($ClientNumber)
	{
		$ClientNumber = (int)$ClientNumber;
		return isset($this->clients_sockets[$ClientNumber]);
	}

	/**
	 * add a socket to the watched sockets
	 * @param resource $Socket
	 * @return void
	 * @access protected
	 */
	protected function WatchSocket($Socket)
	{
		$this->watched_sockets[] = $Socket;
	}
	
	/**
	 * deletes a socket from the watched sockets
	 * @param resource $Socket
	 * @return void
	 * @access private
	 */
	private function UnwatchSocket($Socket)
	{
		$socket_index = array_search($Socket, $this->watched_sockets);
		unset($this->watched_sockets[$socket_index]);
		unset($this->readable_sockets[$socket_index]);
	}

	/**
	 * tells if the socket is in the $this->readable_sockets array. That hopefully means
	 * that the sockets has data to be read
	 * @param resource $Socket
	 * @return bool
	 * @access protected
	 */
	protected function SocketHasData($Socket)
	{
		if ( in_array($Socket, $this->readable_sockets) )
		{
			return TRUE;
		}
		return FALSE;
	}

	/**
	 * return an array of readable client indexes. each index points to some client socket that
	 * has data to be read
	 * @param void
	 * @return array
	 * @access protected
	 */
	protected function GetReadableClients()
	{
		$readable_sockets = array_intersect($this->clients_sockets, $this->readable_sockets);
		$readable_clients = array();
		foreach ( $readable_sockets as $client_index => $client_socket )
		{
			$readable_clients[] = $client_index;
		}
		return $readable_clients;
	}
	
	/**
	 * accepts a client incoming connection
	 * @param void
	 * @return int
	 * @access private
	 */
	private function AcceptConnection()
	{
		$client_socket = socket_accept($this->main_socket);
		/* Making our original address renewable */
		socket_set_option($client_socket, SOL_SOCKET, SO_REUSEADDR, 1);
		/* Getting client index, either from available indexes from previously disconnected clients, or creating new one */
		$client_index = ( count($this->available_indexes) ) ? array_shift($this->available_indexes) : count($this->clients_sockets);
		$this->clients_sockets[$client_index]  = $client_socket;
		/* storing reverse socket/index array for speed and convenience */
		$this->clients_indexes[$client_socket] = $client_index;
		/* Adding it to the watched sockets array */
		$this->WatchSocket($client_socket);
		Logger::WriteLog("Accepting new client # {$client_index} with socket: {$client_socket}");
		$this->OnConnect($client_index);
		return $client_index;
	}

	/**
	 * close the socket connection with some client
	 * @param int $ClientIndex
	 * @return void
	 */
	protected function CloseClientConnection($ClientIndex)
	{
		// Adding to the available indexes array
		array_push($this->available_indexes, $ClientIndex);
		// Getting client socket
		$client_socket = $this->clients_sockets[$ClientIndex];
		// Removing from clients and sockets array
		unset($this->clients_indexes[$client_socket]);
		unset($this->clients_sockets[$ClientIndex]);
		$this->UnwatchSocket($client_socket);
		socket_close($client_socket);
		Logger::WriteLog("Closing client # {$ClientIndex} with socket: {$client_socket}");
		$this->OnDisconnect($ClientIndex);
	}

	/**
	 * sends data to the specified client, it ensures that all the data is sent
	 * if needed, more calls to send data are done recursively
	 * @param int $ClientIndex
	 * @param string $ClientData
	 * @return void
	 */
	protected function SendData($ClientIndex, $ClientData)
	{
		$client_socket = $this->clients_sockets[$ClientIndex];
		$data_length   = strlen($ClientData);
		$wrote_bytes   = socket_write($client_socket, $ClientData, $data_length);
		if ( $wrote_bytes === FALSE )
		{
			Logger::WriteLog("Client {$ClientIndex} seems to be disconnected, closing connection");
			$this->CloseClientConnection($ClientIndex);
			return;
		}
		Logger::WriteLog("Wrote {$wrote_bytes} of {$data_length} bytes", Logger::LOG_LEVEL_DEBUG);
		if ( $wrote_bytes < $data_length )
		{
			$this->SendData($client_socket, substr($ClientData, $data_length)); /* send remaining data */
		}
	}

	/**
	 * reads data from the specified client socket
	 * @param int $ClientIndex
	 * @return string
	 */
	protected function ReadData($ClientIndex)
	{
		$client_socket = $this->clients_sockets[$ClientIndex];
		$client_data   = socket_read($client_socket, SocketLayer::SOCKET_READ_BYTES);
		/* If client data is empty, then it means that client has disconnected so we remove it */
		if ( $client_data === FALSE || $client_data == "" )
		{
			$this->CloseClientConnection($ClientIndex);
			return NULL;
		}
		return $client_data;
	}
	
	/**
	 * start the socket server
	 * @param void
	 * @return void
	 */
	public function Start()
	{
		$this->OnStart();
		while ( $this->WaitForData() )
		{
			$this->OnDataAvailable();
			$readable_clients = $this->GetReadableClients();
			foreach ( $readable_clients as $client_index )
			{
				$client_data = $this->ReadData($client_index);
				if ( $client_data !== NULL )
				{
					$this->OnClientDataReceipt($client_index, $client_data);
				}
			}
		}
	}
	
	/**
	 * this function stops the socket server
	 * @param void
	 * @return void
	 */
	public function Stop()
	{
		socket_close($this->main_socket);
	}
}
?>
